#!/usr/bin/python

from __future__ import division

import sys
import os.path
import os
import math
import lzo
import rfoo
import getopt
import time
from Queue import Queue
import threading
import datetime
import traceback
import zlib
import dioutil
import traceback
import c2padmin

TS_I=0
USER_I=1
NICE_I=2
SYS_I=3
IDLE_I=4
IOWAIT_I=5
IRQ_I=6
SIRQ_I=7

QUEUE_SIZE = 5
NUM_CPU = 8
BLOCK_SIZE = 4 * 1024 * 1024

jiffy = os.sysconf(os.sysconf_names['SC_CLK_TCK'])

def getstats():
	stat_fd = open('/proc/stat')
	stat_buf = stat_fd.readlines()[0].split()
	ret = (time.time(), float(stat_buf[1]), float(stat_buf[2]), float(stat_buf[3]), float(stat_buf[4]), float(stat_buf[5]), float(stat_buf[6]), float(stat_buf[7]))
	stat_fd.close()
	return ret

class Backup_Engine(object):
	def __init__(self, client):
		self._status_file = None
		self._write_queue = Queue(QUEUE_SIZE)
		self._rpc_lock = threading.Lock()
		
		self._time_reading = 0.0
		self._time_CRC = 0.0
		self._time_compresing = 0.0
		self._time_writing = 0.0
		self._time_hardlinking = 0.0
		self._time_waiting_for_queue = 0.0
		self._blocks_linked = 0
		self._blocks_writen = 0
		self._blocks_recompressed = 0
		
		self._start_time = None
		
		self._index_new = None
		self._index_old = None
		
		self._backup_identifier = None
		
		self._client = client
		self._name = None
		self._type = None
		self._name = None
		
		self._src = None
		self._src_lv = None
		self._last_block = None
		self._block_range = []
		
		self._backup_identifier = None
		
		self._write_worker_thread = None
		self._write_worker_connected = None
		self._write_worker_thread_running = False
		self._run = False
		self._complete = False
		self._cancel = False
		self._origin = False
		self._last_activity = 0
		
		self._result = [0, 'init']
		self._job_id = None
		# result - [error_code, text_status]
		# 0, status - OK/Cancel
		# 1, breaking - backup break correctly
		# 2, write_worker_end - connection lost
		# 3, write_worker_end - write worker exit unexpectedly
		
		os.system('ionice -c3 -p%s >/dev/null 2>&1' % (os.getpid()))
		
	def _write_status(self, block_number, progress = None):
		fh = open(self._status_file, 'w')
		fh.write("%d\n" % int(block_number * BLOCK_SIZE))
		fh.close()
		self._client.update_backup_progress(self._backup_identifier, [int(block_number), int(self._last_block)])
		self._last_activity = time.time()
			
		if progress != None:
			progress.set(float(block_number) / float(self._last_block))
		
	def start_backup(self, name, src_lv, btype, ts = None, orig_name = None, metadata = {}, job_id = None):
		if(self._write_worker_thread != None and self._write_worker_thread_running):
			raise Exception('Couldn\'t start backup twice')
		
		self._name = name
		self._src_lv = src_lv
		if(job_id == None):
			self._job_id = os.urandom(8).encode('hex')
			
		else:
			self._job_id = job_id	
			
		
		self._status_file = '/dev/shm/vps-backup-process-status-%s' % (self._job_id,)	
		if(self._src_lv == '-'):
			self._src = sys.stdin
			
		else:	
			self._src = dioutil.DioFile()
			self._src.open(self._src_lv)
			lb = self._src.getsize()
			self._src.seek(0)
			self._last_block = int(math.ceil(lb / BLOCK_SIZE)) - 1
			self._block_range = range(self._last_block + 1)
			
		self._type = btype
		
		metadata['__block_size__'] = BLOCK_SIZE
		self._backup_identifier, self._index_old, self._origin = self._client.startBackup(self._name, self._type, ts, orig_name, metadata, self._job_id)
		self._last_activity = time.time()
		self._index_new = {}
		self._size = 0
		
		self._write_worker_thread = threading.Thread(target=self._write_worker)
		self._write_worker_thread.daemon = True
		
		self._result = [0, 'backup prepared']
		
		self._run = False
		self._complete = False
		self._cancel = False
		
		return self._backup_identifier, self._origin
	
	def resume_backup(self, name = None, src_lv = None, btype = None, backup_identifier = None, origin = False, index = None, job_id = None, client = None):
		if(self._write_worker_thread != None and self._write_worker_thread_running):
			raise Exception('Couldn\'t resume backup twice')
			
		if(name != None):
			self._name = name
		
		if(src_lv != None):
			self._src_lv = src_lv
			
		if(backup_identifier != None):
			self._backup_identifier = backup_identifier
			
		if(origin != False):
			self._origin = origin
			
		if(index != None):
			self._index_new = index
			
		if(client != None):
			self._client = client
			
		if(btype != None):
			self._type = btype
			
		if(job_id != None):
			self._job_id = job_id
		
		if(self._name == None):
			raise Exception('Unknown backup source name')
			
		if(self._src_lv == None):
			raise Exception('Unknown backup source')
			
		if(self._src_lv == '-'):
			raise Exception('Couldn\'t resume on stdin')
			
		if(self._backup_identifier == None):
			raise Exception('Unknown backup for continue')
			
		if(self._origin == False):
			raise Exception('Unknown origin')
			
		if(self._index_new == None):
			raise Exception('Unknown previous backup status')
			
		if(self._client == None):
			raise Exception('Backup client is not set')
			
		if(self._type == None):
			raise Exception('Backup type is not set')
			
		if(self._job_id == None):
			raise Exception('Job ID is not set')
		
		self._status_file = '/dev/shm/vps-backup-process-status-%s' % (self._job_id,)
		self._src = dioutil.DioFile()
		self._src.open(self._src_lv)
		lb = self._src.getsize()
		self._src.seek((len(self._index_new) - 1) * BLOCK_SIZE)
		self._last_block = int(math.ceil(lb / BLOCK_SIZE)) - 1
		self._block_range = range(len(self._index_new) - 1, self._last_block + 1)
	
		self._backup_identifier, self._index_old, self._origin = self._client.resumeBackup(self._name, self._backup_identifier, self._origin)
		self._last_activity = time.time()
		self._size = (len(self._index_new) - 1) * BLOCK_SIZE
		
		self._write_worker_thread = threading.Thread(target=self._write_worker)
		self._write_worker_thread.daemon = True
		
		self._result = [0, 'backup prepared']
		
		self._run = False
		self._complete = False
		self._cancel = False
		
		return self._backup_identifier, self._origin
	
	def run(self, progress = None):
		if(self._write_worker_thread == None):
			raise Exception('Couldn\'t run before initialization')
			
		if(self._run):
			raise Exception('Couldn\'t run backup twice')
		
		self._run = True
		self._write_worker_thread.start()
		
		now = time.time()
		if(self._start_time == None):
			self._start_time = now
			
		stats_time = now + 10
		my_iowait = 0.0
		read_times = []
		stats = getstats()
		
		self._result = [0, 'backup_start']
		if(self._src_lv == '-'):
			block_number = 0
			while True:
				if block_number % 10 == 0: 
					self._write_status(block_number)
				
				if(self._last_activity + 5 < time.time()):
					self._client.ping()
					self._last_activity = time.time()
				
				now = time.time()
				src_block = self._src.read(BLOCK_SIZE)
				self._size += len(src_block)
				self._time_reading += time.time() - now
				if(src_block == ''):
					break
				
				self._write_queue.put((block_number, src_block))
				block_number += 1
			
		else:
			for block_number in self._block_range:
				if(not self._run):
					self._result = [1, 'breaking']
					break
				
				if(block_number % 10 == 0): 
					try:
						self._write_status(block_number, progress)
						
					except Exception as ex:
						print ex
						self._result = [2, 'connection lost']
						self._run = False
						break
				
				now = time.time()
				if(now > stats_time):
					stats_time = now + 10
					stats_new = getstats()
					my_iowait = (stats_new[IOWAIT_I] - stats[IOWAIT_I]) / jiffy / NUM_CPU / (stats_new[0] - stats[0])
					stats = stats_new
			
				src_block = self._src.read(BLOCK_SIZE)
				if block_number == self._last_block:
					self._size += len(src_block)
				
				else:
					self._size += BLOCK_SIZE
				
				read_time = time.time() - now
				read_times.append(read_time)
			
				if len(read_times) > 10:
					read_times.pop(0)
				
				sleep_time = (my_iowait ** 3) * 5 * (sum(read_times) / len(read_times))
				if sleep_time > 0.3:
					sleep_time = 0.3
				
				self._time_reading += read_time
				if(not src_block):
					raise IOError('Read error when processing block %u of %u.' % (block_number, self._last_block))
			
		
				now = time.time()
				while True:
					try:
						self._write_queue.put((block_number, src_block), True, 10)
						break
					
					except:
						if(not self._run):
							break
						
						if(not self._write_worker_thread.is_alive()):
							if(self._write_worker_connected):
								self._result = [3, 'write_worker_end']
							
							else:
								self._result = [2, 'write_worker_lost_connection']
							
							self._src.close()
							return 
						
						else:
							continue
						
				waited = time.time() - now
				self._time_waiting_for_queue += waited
				if (sleep_time - waited) > 0.001:
					time.sleep(sleep_time - waited)
		
		if(self._result[0] == 0):
			self._result = [0, 'read_complete']
			
		self._complete = self._run
			
		if(self._src_lv != '-'):
			self._src.close()
			
		self._write_worker_thread.join()
		self._run = False
		if(self._result[0] == 0):
			self._result = [0, 'backup_complete']
		
		if(self._complete):
			self._write_index()
			
		if(self._cancel):
			self._write_index()
			self._clean()
			self._result = [0, 'backup_cancel']
			
		time_total = time.time() - self._start_time
		try:
			if os.path.exists('/tmp/backup-stats-' + name):
				fh = open('/tmp/backup-stats-'+name, 'a')
						
			else:
				fh = open('/tmp/backup-stats-'+name, 'w')
				fh.write('total	reading	CRC	compressing	writing	linking	blocksLinked	blocksWriten	waitedForQueue	blocksRecompressed\n')
			
			fh.write('%s	%s	%s	%s	%s	%s	%s	%s	%s	%s\n' % (timeTotal, self._time_reading, self._time_CRC, self._time_compresing, self._time_writing, self._time_hardlinking, self._blocks_linked, self._blocks_writen, self._time_waiting_for_queue, self._blocks_recompressed))
			fh.close()
			
		except:
			pass
		
		try:
			os.unlink(self._status_file)
			
		except:
			pass
		
		return
	
	def stop(self):
		self._run = False
		return True
		
	def cancel(self):
		self._run = False
		self._cancel = True
		return True
	
	def _write_worker(self):
		self._write_worker_thread_running = True
		zero_block = chr(0) * BLOCK_SIZE
		self._write_worker_connected = True
		try:
			while self._run:
				if(self._last_activity + 5 < time.time()):
					self._client.ping()
					self._last_activity = time.time()
				
				try:
					block_number, src_block = self._write_queue.get(True, 1)
			
				except:
					if self._complete:
						break
				
					else:
						continue
					
				index_new = [0,0.0,'']		
				now = time.time()
				if(src_block == zero_block): # When current block is NULL, write only to index
					index_new[0] = -1
					self._index_new[block_number] = index_new
					continue
				
				index_new[0] = lzo.adler32(src_block)
				self._time_CRC += time.time() - now
			
				if(not self._index_old.has_key(block_number)):
					self._index_old[block_number] = [-1,time.time(),'lzo']
				
				now = time.time()
			
				# rovna se nam hash a zaroven mame budto vysokou uroven komprese, nebo je blok mladej (a zaroven nemame manualni zalohu nebo image)
				try:
					if(index_new[0] == self._index_old[block_number][0] and (self._index_old[block_number][2] != 'lzo' or (self._type not in ['man', 'image'] and (now - self._index_old[block_number][1]) < 432000)) and self._hardlink_block(block_number)):
						self._time_hardlinking += time.time() - now
						self._blocks_linked += 1
						# nalinkovani bloku jsem pridal uz do podminky, protoze kdyz selze, blok budeme chtit normalne zapsat
						self._index_new[block_number] = self._index_old[block_number]
					else:
						if(self._type in ['man', 'image'] or (index_new[0] == self._index_old[block_number][0] and self._index_old[block_number][2] == 'lzo' and now - self._index_old[block_number][1]) >= (432000)):
							index_new[1] = self._index_old[block_number][1]
							index_new[2] = 'zlib'
							now = time.time()
							dst_block = zlib.compress(src_block)
							self._time_compresing += time.time() - now
							self._blocks_recompressed += 1
					
						else:
							index_new[1] = time.time()
							index_new[2] = 'lzo'
							st = time.time()
							dst_block = lzo.compress(src_block)
							self._time_compresing += time.time() - st
				
						now = time.time()
						self._write_block(block_number, dst_block)
						self._time_writing += time.time() - now
						self._blocks_writen += 1
						self._index_new[block_number] = index_new
				
				except:
					traceback.print_exc()
					self._write_worker_connected = False				#connection lost
					break
					
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
			
		self._write_worker_thread_running = False

	
	def _write_block(self, block_number, dst_block):
		with self._rpc_lock:
			ret = self._client.writeBlock(self._name, self._backup_identifier, block_number, dst_block)
			self._last_activity = time.time()
		
		return ret

	def _hardlink_block(self, block_number):
		with self._rpc_lock:
			ret = self._client.linkBlock(self._name, self._backup_identifier, block_number)
			self._last_activity = time.time()
		
		return ret

	def _write_index(self):
		return self._client.finishBackup(self._name, self._backup_identifier, self._index_new, self._size)
		
	def _clean(self):
		return self._client.removeBackup(self._name, self._backup_identifier)
		
	def get_index_new(self):
		return self._index_new
		
	def get_result(self):
		return self._result
		
	def get_write_status(self):
		return [len(self._index_new) if(self._index_new != None) else None, (self._last_block + 1) if(self._last_block != None) else None]
