import argparse,sys,time,traceback
from c2padmin.backups.client import BackupClient
from c2padmin.backups import backup,restore

def run(in_args):
	"""docstring for run"""
	global _client
	parser = argparse.ArgumentParser()

	parser.add_argument('-a','--archive',help='Archive address',default=None)
	parser.add_argument('--public',help='Use public domain access',default=False,const=True,action='store_const')
	parser.add_argument('-j', '--action',required=True,choices=['list','list_all','backup','restore','delete'], help='What should we do')
	parser.add_argument('-n', '--name',required=True,help='Backup module name e.g. virtual_123')
	parser.add_argument('-p','--path',help='Backup path / Backup identifier')
	parser.add_argument('-t','--type',help='Type of newly created backup',choices=['auto','man','image'],default='man')
	parser.add_argument('-d','--device',help='Backup this device / restore to this device. Use "-" for stdin/stdout')
	parser.add_argument('-o','--origin',help='Backup module of backup orgin (use this to backup data which is similar to another pre-existing backup)')
	parser.add_argument('-c','--console',help='Force console type output even if stdout is not a tty',action='store_const',const=True,default=sys.stdout.isatty())
	parser.add_argument('-f','--force',help='Force rewrite data',default=False,const=True,action='store_const')
	parser.add_argument('-s','--timestamp',help='New backup custom timestamp')
	args = parser.parse_args(in_args)
	
	_client = BackupClient()
	_client.connect(backup_host=args.archive,public=args.public)
	
	if args.action == 'list':
		list_backups(args.name,args.console)
		
	elif args.action == 'list_all':
		list_backups(args.name,args.console, True)
		
	elif args.action == 'delete':
		if args.path == None:
			parser.error('Backup path not specified')
		remove_backup(args.name,args.path)
	elif args.action == 'backup':
		if args.device == None:
			parser.error('Backup device not specified')
		if args.timestamp != None:
			try:
				ts = int(time.mktime(datetime.datetime(*(time.strptime(args.timestamp, '%Y-%m-%d-%H-%M-%S')[0:6])).timetuple()))
			except:
				ts = None
		else:
			ts = None
		try:
			backup_engine = backup.Backup_Engine(_client)
			backup_engine.start_backup(args.name, args.device, args.type, ts, args.origin)
			print backup_engine.run()

		except:
			sys.stderr.write(traceback.format_exc())
			sys.stderr.flush()
	elif args.action == 'restore':
		if args.device == None:
			parser.error('Backup device not specified')
		if args.path == None:
			parser.error('Backup path not specified')
		available_backups = _client.listBackups(args.name)
		backup_found = False
		for item in available_backups:
			if args.path == item['path']:
				backup_found = True
				break
		if not backup_found:
			parser.error('Backup specified by path does not exist')
		try:
			restore_engine = restore.Restore_Engine(_client)
			restore_engine.start_restore(args.name, args.device, args.path)
			restore_engine.run(not args.force)
			
		except:
			sys.stderr.write(traceback.format_exc())
			sys.stderr.flush()
	else:
		assert False, 'Unhandled action %s' % (args.action,)


def list_backups(name,console=False, with_running = False):
	if(name == '-'):
		name = None
		
	ret = _client.listBackups(name, with_running)
	if console:
		if(with_running):
			if(len(ret[0]) == 0 and len(ret[1]) == 0):
				return
				
			line = '%s	%s	%s	%s	%s	%s'
			for item in ret[0]:
				if(name == None):
					subject = item['path'].split('-',1)[0]
				
				else:
					subject = name
					
				print line % (subject, item['path'], item['type'], item['time'], item['size'], 'complete')
				
			for item in ret[1]:
				if(name == None):
					subject = item['path'].split('-',1)[0]
				
				else:
					subject = name
					
				print line % (subject, item['path'], item['type'], item['time'], '    ---    ', 'running')
		
		else:
			if(len(ret) == 0):
				return
				
			for item in ret:
				if(name == None):
					subject = item['path'].split('-',1)[0]
				
				else:
					subject = name
					
				print '%s	%s	%s	%s	%s' % (subject, item['path'], item['type'], item['time'], item['size'])
				
	else:
		sys.stdout.write(rfoo.marsh.dumps(ret))

def remove_backup(name, path):
	print _client.removeBackup(name, path)	
	