import rfoo,ssl,threading

class BackupClient(object):
	"""docstring for BackupClient"""
	def __init__(self):
		super(BackupClient, self).__init__()
		self._conn = None
		self._api = None
		self._communication_lock = threading.Lock()

	def connect(self, backup_host=None,backup_port=None, public = False, timeout = None):
		"""docstring for connect"""
		from c2padmin import config
		with self._communication_lock:
			if backup_host == None:
				backup_host = config.config.get('node','backup_host')
			
			if backup_port == None:
				backup_port = config.config.get('node','backup_port',int)
			
			if(timeout == None):
				timeout = config.config.get('node', 'backup_timeout', int)
		
			ssl_context = ssl.create_default_context(cafile=config.config.get('node','ca_bundle'))
			ssl_context.check_hostname = False
			if not public:
				ssl_context.load_cert_chain(certfile=config.config.get('node','backup_cert'), keyfile=config.config.get('node','backup_key'))
			
			else:
				if(config.config.get('node','backup_public_cert') != '' and config.config.get('node','backup_public_key') != ''):
					passwd = config.config.get('node','backup_public_password')
					passwd = None if(passwd == '') else passwd
					ssl_context.load_cert_chain(certfile=config.config.get('node','backup_public_cert'), keyfile=config.config.get('node','backup_public_key'), password = passwd)
							
			self._conn = rfoo.SSLConnection(ssl_context, timeout = timeout).connect(host=backup_host, port=backup_port)
			self._api = rfoo.Proxy(self._conn)
	
	def disconnect(self):
		with self._communication_lock:
			self._conn.close()
			self._api = None
	
	def ping(self):
		"""docstring for ping"""
		with self._communication_lock:
			return self._api.ping()
		
	def getAccount(self):
		"""docstring for ping"""
		with self._communication_lock:
			return self._api.getAccount()
		
	def getLastAccess(self):
		"""docstring for getLastAccess"""
		with self._communication_lock:
			return self._api.getLastAccess()
		
	def listSubjects(self):
		with self._communication_lock:
			return self._api.listSubjects()
			
	def listBackups(self, subject, with_running = False):
		with self._communication_lock:
			return self._api.listBackups(subject,with_running)
	
	def startBackup(self, subject, btype = None, ts = None, original_subject = None, metadata = {}, job_id = None):
		with self._communication_lock:
			return self._api.startBackup(subject, btype, ts, original_subject, metadata, job_id)
		
	def resumeBackup(self, subject, backup_identifier, origin):
		with self._communication_lock:
			return self._api.resumeBackup(subject, backup_identifier, origin)
	
	def update_backup_progress(self, backup_identifier, progress):
		with self._communication_lock:
			return self._api.update_backup_progress(backup_identifier, progress)
	
	def finishBackup(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.finishBackup(*args,**kwargs)
		
	def startRestore(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.startRestore(*args,**kwargs)
		
	def update_restore_progress(self,backup_identifier, job_id, progress):
		with self._communication_lock:
			return self._api.update_restore_progress(backup_identifier, job_id, progress)
		
	def finishRestore(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.finishRestore(*args,**kwargs)
		
	def writeBlock(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.writeBlock(*args,**kwargs)
		
	def linkBlock(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.linkBlock(*args,**kwargs)
	
	def readBlock(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.readBlock(*args,**kwargs)
	
	def removeBackup(self,*args,**kwargs):
		with self._communication_lock:
			return self._api.removeBackup(*args,**kwargs)
		
	def change_backup_type(self, subject, backup_identifier, btype = None):
		with self._communication_lock:
			return self._api.change_backup_type(subject, backup_identifier, btype)
		
	def change_metadata(self, backup_identifier, data):
		with self._communication_lock:
			return self._api.change_metadata(backup_identifier, data)
		