import time,threading,Queue,pickle

from c2padmin import logger as logmodule
from c2padmin.commands import CommandProgress
from types import BackupSchedule,BackupJob,Restore_Job,Job_Queue
from client import BackupClient
from c2padmin.virt.guest import get_name_from_uuid

import client
import c2padmin

logger = logmodule.getLogger(__name__)


MAX_RUNNING_BACKUPS=2
MAX_RUNNING_RESTORES=2

class BackupDispatcher(object):
	"""docstring for BackupRunner"""
	def __init__(self):
		super(BackupDispatcher, self).__init__()
		#self._queue = Queue.Queue()
		self._queue = Job_Queue()
		self._remove_queue = Queue.Queue()		
		self._pending = {} # list of VPS names being processed
		self._threads = []
		ret = self.load()
		
		
	def cancel(self, backup_identifier):
		for mythread in self._threads:
			if(mythread.get_backup_identifier() == backup_identifier):
				mythread.cancel()
				logger.info('Backup thread {} canceled'.format(mythread.get_name()))
		
	def main_loop(self):
		"""docstring for main_loop"""
		try:
			while True:
				if c2padmin.config.main_thread_signal != None:
					for mythread in self._threads:
						mythread.stop()
						mythread.join()
						self._queue.put(mythread.get_copy())
				
					self.save()
					return
				
				for mythread in self._threads:
					if not mythread.is_alive():
						uuid = mythread.get_uuid()
						try:
							domname = get_name_from_uuid(uuid)
						except:
							domname = mythread.get_name()
					
						try:
							self._threads.remove(mythread)
						
						except:
							pass
						
						result = mythread.get_result()
						if(type(mythread) == BackupJob):
							if(result[0] == 0):
								logger.info('Backup thread %s finished successfully' % (mythread.get_name(),))
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
					
							elif(result[0] == 1):
								logger.error('Backup thread %s failed with uncaught exception: %s' % (mythread.get_name(), result[1]))	
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
					
							elif(result[0] == 2):
								logger.error('Backup thread %s failed: %s' % (mythread.get_name(), result[1]))
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
						
							elif(result[0] == 3):
								job = mythread.get_copy()
								self._queue.put(job)
								self._pending[domname] = job
								logger.warning('Backup server is unavailable')
								with c2padmin.config.node_config_lock:
									if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] == None):
										c2padmin.config.node_config['backup_server']['unavailable_timestamp'] = time.time()
						
							elif(result[0] == 4):
								job = mythread.get_copy()
								self._queue.put(job)
								self._pending[domname] = job
								logger.warning('Backup break during initialize backup')
								with c2padmin.config.node_config_lock:
									if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] == None):
										c2padmin.config.node_config['backup_server']['unavailable_timestamp'] = time.time()
								
							elif(result[0] == 5):
								job = mythread.get_copy()
								self._queue.put(job)
								self._pending[domname] = job
								logger.warning('Backup break during backup')
								with c2padmin.config.node_config_lock:
									if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] == None):
										c2padmin.config.node_config['backup_server']['unavailable_timestamp'] = time.time()
					
							elif(result[0] == 6):
								logger.warning('Backup end during finalize backup (node_config could be broken)')
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
						
							else:
								print result
							
						else:
							if(result[0] == 0):
								logger.info('Restore thread %s finished successfully' % (mythread.get_name(),))
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
					
							elif(result[0] == 1):
								logger.error('Restore thread %s failed with uncaught exception: %s' % (mythread.get_name(), result[1]))	
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
					
							elif(result[0] == 2):
								logger.error('Restore thread %s failed: %s' % (mythread.get_name(), result[1]))
								try:
									del self._pending[domname]
					
									if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
										with c2padmin.config.node_config_lock:
											c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = None
							
								except:
									logger.warning('Couldn\'t remove backup/restore from pendings.')
						
							elif(result[0] == 3):
								job = mythread.get_copy()
								self._queue.put(job)
								self._pending[domname] = job
								logger.warning('Backup server is unavailable')
								with c2padmin.config.node_config_lock:
									if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] == None):
										c2padmin.config.node_config['backup_server']['unavailable_timestamp'] = time.time()
						
							elif(result[0] == 4):
								job = mythread.get_copy()
								self._queue.put(job)
								self._pending[domname] = job
								logger.warning('Restore break during initialize restore')
								with c2padmin.config.node_config_lock:
									if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] == None):
										c2padmin.config.node_config['backup_server']['unavailable_timestamp'] = time.time()
								
							elif(result[0] == 5):
								job = mythread.get_copy()
								self._queue.put(job)
								self._pending[domname] = job
								logger.warning('Restore break during restore')
								with c2padmin.config.node_config_lock:
									if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] == None):
										c2padmin.config.node_config['backup_server']['unavailable_timestamp'] = time.time()
						
							else:
								print result
				
					
				if(c2padmin.config.node_config['backup_server']['unavailable_timestamp'] != None):
					time.sleep(0.1)
					continue
			
				backup_count = len(filter(lambda x: type(x) == BackupJob, self._threads))
				restore_count = len(self._threads) - backup_count
			
				if(backup_count < MAX_RUNNING_BACKUPS):
					try:
						mythread = self._queue.get('backup')
						self._threads.append(mythread)
						mythread.start()
						logger.debug('Backup thread %s started' % (mythread.get_name(),))
					
					except Queue.Empty:
						pass
					
				if(restore_count < MAX_RUNNING_RESTORES):
					try:
						mythread = self._queue.get('restore')
						self._threads.append(mythread)
						mythread.start()
						logger.debug('Restore thread %s started' % (mythread.get_name(),))
					
					except Queue.Empty:
						pass
					
				try:
					to_remove = self._remove_queue.get(False)
					complete, running = self.backup_list()
					complete = filter(lambda x: x['path'].split('-')[0] == to_remove['name'], complete)
					running = filter(lambda x: x['path'].split('-')[0] == to_remove['name'], running)
					if('backup_identifiers' in to_remove and type(to_remove['backup_identifiers']) == list):
						complete = filter(lambda x: x['path'] in to_remove['backup_identifiers'], complete)
						running = filter(lambda x: x['path'] in to_remove['backup_identifiers'], running)
					
					elif('type' in to_remove):
						if(to_remove['type'] != None):
							complete = filter(lambda x: x['type'] == to_remove['type'], complete)
							running = filter(lambda x: x['type'] == to_remove['type'], running)
					
					else:
						logger.warning('Bad data format in remove queue')
				
					for backup in complete:
						self._remove_backup(to_remove['name'], backup['path'])
						logger.info('Backup {} removed'.format(backup['path']))
			
					for backup in running:
						self.cancel(backup['path'])
						logger.info('Backup {} canceled'.format(backup['path']))
				
				except Queue.Empty:
					pass
			
				time.sleep(0.1)
				
		except Exception as ex:		
			c2padmin.dispatcher.unexpectly_end(ex)
	
	def backup_list(self):
		client = BackupClient()
		client.connect()
		ret = client.listBackups(None, True)
		client.disconnect()
		return ret
			
	def save(self):
		data = []
		while True:
			try:
				item = self._queue.get()
				
			except Queue.Empty:
				break
			
			data.append(item)
			
		c2padmin.cache.cache.save_config('backup_queue', data)
		
		data = []
		while True:
			try:
				item = self._remove_queue.get(False)
				
			except Queue.Empty:
				break
			
			data.append(item)
		
		c2padmin.cache.cache.save_config('backup_remove_queue', data)
		
		
	def load(self):
		retVal = False
		
		data = c2padmin.cache.cache.load_config('backup_queue')
		if(data != False and data != None):
			with c2padmin.config.node_config_lock:
				for job in data:
					if(type(job) == BackupJob):
						snapshot = job.get_snapshot()
						if(snapshot != None and snapshot not in c2padmin.config.node_config['active_snapshots']):
							c2padmin.config.node_config['active_snapshots'][snapshot] = time.time()
						
					uuid = job.get_uuid()
					
					if(uuid != None and uuid in c2padmin.config.node_config['libvirt']['domains']):
						self._queue.put(job)
						domname = get_name_from_uuid(uuid)
						self._pending[domname] = job
						c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = job.get_job_id()
						
					else:
						logger.error('Scheduled backup/restore couldn\'t be processed, domain {} doesn\'t exists'.format(uuid))
						
					
				
			c2padmin.cache.cache.commit_config('backup_queue')
			retVal = True
			
		elif(data != None):
			logger.error('Couldn\'t restore backup_queue')
			
		data = c2padmin.cache.cache.load_config('backup_remove_queue')
		if(data != False and data != None):
			for job in data:
				self._remove_queue.put(job)
				
			c2padmin.cache.cache.commit_config('backup_remove_queue')
			retVal = retVal and True
			
		elif(data != None):
			logger.error('Couldn\'t restore backup_remove_queue')
			
		return retVal
		
	def get_pending(self):
		return self._pending	
			
	def scheduler_loop(self):
		"""docstring for scheduler_loop"""
		try:
			while True:
				if c2padmin.config.main_thread_signal != None:
					# TODO
					return
				with c2padmin.config.node_config_lock:
					for domain in c2padmin.config.node_config['libvirt']['domains']:
						try:
							domname = get_name_from_uuid(domain)
						except:
							continue
					
						try:
							if(c2padmin.config.node_config['libvirt']['domains'][domain]['backups']['schedule'] != None):
								c2padmin.config.node_config['libvirt']['domains'][domain]['backups']['schedule'].delete_old()
								if(c2padmin.config.node_config['libvirt']['domains'][domain]['backups']['schedule'].backup_now() and domname not in self._pending):
									job = BackupJob(domname)
									self._queue.put(job)
									self._pending[domname] = job
									logger.info('Scheduled backup of {}'.format(domname))
							
						except:
							pass
					
				time.sleep(1)
			
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
	
	def remove_backups(self, name, backup_identifiers = [], btype = None):
		if(btype != None and btype != 'man' and btype != 'auto' and btype != 'image'):
			raise Exception('Bad backup type')
		
		if(len(backup_identifiers) != 0):	
			self._remove_queue.put({'name': name, 'backup_identifiers': backup_identifiers})
			
		else:
			self._remove_queue.put({'name': name, 'type': btype})
			
		return True
		
	def _remove_backup(self, name, backup_identifier):
		client = BackupClient()
		client.connect()
		ret = client.removeBackup(name, backup_identifier)
		client.disconnect()
		return True
			
	def make_backup(self, uuid, metadata = {}, snapshot = None, image = False):
		if(uuid not in c2padmin.config.node_config['libvirt']['domains']):
			logger.warning('Domain %s not found' % (uuid,))
			raise Exception('Domain %s not found' % (uuid,))
			
		domname = get_name_from_uuid(uuid)
		if(domname not in self._pending):
			if(image):
				job = BackupJob(domname, 'image', 0, metadata, snapshot, uuid = uuid)
				
			else:
				job = BackupJob(domname, 'man', 0, metadata, snapshot, uuid = uuid)
			
			self._queue.put(job)	
			self._pending[domname] = job
			with c2padmin.config.node_config_lock:
				c2padmin.config.node_config['libvirt']['domains'][uuid]['backups']['pending'] = job.get_job_id()
				
			logger.info('Backup for server %s will be processed' % (domname,))
			return job.get_job_id()
		
		logger.warning('Backup/restore for server %s is pending' % (domname,))
		raise Exception('Backup/restore for server %s is pending' % (domname,))
		
	def make_restore(self, uuid, name, backup_identifier, dst_lv, public = False, progress = None):
		if(uuid not in c2padmin.config.node_config['libvirt']['domains']):
			logger.warning('Domain %s not found' % (uuid,))
			raise Exception('Domain %s not found' % (uuid,))
			
		domname = get_name_from_uuid(uuid)
		if(domname not in self._pending):
			job = Restore_Job(name, backup_identifier, dst_lv, public, uuid = uuid, disallow_discard = c2padmin.config.config.get('node', 'disallow_discard', bool))
			self._queue.put(job)
			self._pending[domname] = job
			logger.info('Restore of server %s will be processed' % (domname,))
			while(job in self._threads or domname in self._pending):
				try:
					job = self._pending[domname]
				
				except:
					
					continue
					
				status = job.get_restore_status()
				if(progress != None):
					if(type(status[0]) == int and type(status[1]) == int):
						progress.set(float(status[0]) / float(status[1]))
						
					else:
						progress.set(0)
						
				time.sleep(1)
			
			return job.get_result()
		
		else:
			job = self._pending[domname]
			if(type(job) == Restore_Job):
				while(job in self._threads or domname in self._pending):
					try:
						job = self._pending[domname]
				
					except:
						continue
						
					status = job.get_restore_status()
					if(progress != None):
						if(type(status[0]) == int and type(status[1]) == int):
							progress.set(float(status[0]) / float(status[1]))
						
						else:
							progress.set(0)
							
					time.sleep(1)
			
				return job.get_result()
		
		logger.warning('Backup/restore of server %s is pending' % (domname,))
		raise Exception('Backup/restore of server %s is pending' % (domname,))
		
		
	
	def worker(self):
		"""docstring for worker"""
		pass
		
	def get_running_jobs(self):
		"""docstring for get_running_jobs"""
		pass