import rfoo,ssl,threading,os,copy,threading,time,ConfigParser, Queue

from c2padmin import virt
from c2padmin import logger as logmodule

import client,backup,restore

import datetime

import c2padmin

logger = logmodule.getLogger(__name__)

FLAG_POWEROFF=1
FLAG_NO_SNAPSHOT=2

class Job_Queue(object):
	def __init__(self):
		self._backup_queue = Queue.Queue()
		self._restore_queue = Queue.Queue()
		
	def put(self, item):
		if(type(item) == BackupJob):
			self._backup_queue.put(item)
			
		elif(type(item) == Restore_Job):
			self._restore_queue.put(item)
			
		else:
			raise TypeError('can put only backup or restore job')
			
	def get(self, qtype = None):
		if(qtype == None):
			try:
				item = self._backup_queue.get(block = False)
				return item
				
			except:
				pass
				
			item = self._restore_queue.get(block = False)
			return item
			
		elif(qtype == 'backup'):
			item = self._backup_queue.get(block = False)
			return item
			
		elif(qtype == 'restore'):
			item = self._restore_queue.get(block = False)
			return item
			
		else:
			raise ValueError('bad queue type')
			

class BackupList(object):
	"""docstring for BackupList"""
	def __init__(self,name, public = False, init_list = []):
		super(BackupList, self).__init__()
		self._backup_list = init_list
		self._name = name
		self._public = public
		self._lock = threading.Lock()
		self._available = self.refresh()
	
	def __reduce__(self):
		return (self.__class__,(self._name, self._public, self._backup_list))
	
	def is_available(self):
		return self._available
	
	def refresh(self):
		"""docstring for refresh"""
		try:
			_client = client.BackupClient()
			_client.connect(public = self._public)
			with self._lock:
				self._backup_list = _client.listBackups(self._name)
				
		except:
			return False
			
		return True
		
	def get_list(self, metadata_filter = None):
		"""docstring for get_list"""
		with self._lock:
			res = copy.deepcopy(map(lambda x: dict(x,**{'name':self._name}), sorted(self._backup_list,key=lambda x:x['time'])))
			if(metadata_filter != None and type(metadata_filter) == list):
				for index in range(len(res)):
					for key in res[index]['metadata'].keys():
						if(key not in metadata_filter):
							del res[index]['metadata'][key]
						
			return res
	
	def get_last_backup(self, btype = None):
		"""docstring for get_last_backup"""
		if(btype == None):
			ret = self.get_list()
			if len(ret) == 0:
				return None
				
			else:
				return ret[-1]
				
		else:
			ret = filter(lambda x: x['type'] == btype, self.get_list())
			if len(ret) == 0:
				return None
				
			else:
				return ret[-1]
				
class Restore_Job(threading.Thread):
	def __init__(self, name, backup_identifier, dst_lv, public = False, prev_index = {}, restore_engine = None, job_id = None, uuid = None, disallow_discard = False):
		assert type(name) == str, "Invalid name type"
		if(job_id == None):
			self._job_id = os.urandom(8).encode('hex')
			
		else:
			self._job_id = job_id
	
		super(Restore_Job, self).__init__(name='restore_thread_{}_{}'.format(name,self._job_id))
		
		self._name = name
		self._backup_identifier = backup_identifier
		self._dst_lv = dst_lv
		self._public = public
		self._index = prev_index
		self._restore_engine = restore_engine
		self._uuid = uuid
		self._disallow_discard = disallow_discard
		
		self._result = [0, 'init']
		# result - [error_code, text_status]
		# 0, status - OK
		# 1, exception - Uncaught exception
		# 2, reason - Restore couldn't be done
		# 3, reason - Couldn't connect to server
		# 4, reason - Error during restore init on server side
		# 5, reason - Restore brake during process save data to continue
		# 6, reason - Error occured during finalize restore
		
	def __reduce__(self):
		return (self.__class__, (self._name, self._backup_identifier, self._dst_lv, self._public, self._index, None, self._job_id, self._uuid, self._disallow_discard))
	
	def get_copy(self):
		retVal = self.__class__(self._name, self._backup_identifier, self._dst_lv, self._public, self._index, self._restore_engine, self._job_id, self._uuid, self._disallow_discard)
		return retVal
		
	def stop(self):
		self._restore_engine.stop()
		
	def get_backup_identifier(self):
		return self._backup_identifier
		
	def get_name(self):
		return self._name
	
	def get_result(self):
		return self._result
		
	def get_restore_status(self):
		return self._restore_engine.get_restore_status() if self._restore_engine != None else [None, None]
		
	def get_job_id(self):
		return self._job_id
		
	def get_uuid(self):
		return self._uuid
		
	def run(self):
		try:
			if(len(self._index) == 0):
				logger.info('Started restore of {}'.format(self._name))
				
			else:
				logger.info('Started restore of {} from block {}'.format(self._name, len(self._index)))
				
			self._result = [0, 'prepared']
				
			self._client = client.BackupClient()
			
			try:
				self._client.connect(public = self._public)
				self._result = [0, 'connected']
				
			except Exception as ex:
				logger.error('Couldn\'t connect to backup server')
				self._result = [3, ex.message]
				return
				
				
			try:
				if(self._restore_engine == None):	
					self._restore_engine = restore.Restore_Engine(self._client)
					if(len(self._index) == 0):
						self._restore_engine.start_restore(self._name, self._dst_lv, self._backup_identifier, job_id = self._job_id, disallow_discard = self._disallow_discard)
						self._result = [0, 'restore_start']
						
					else:
						self._restore_engine.resume_restore(self._name, self._dst_lv, self._backup_identifier, self._index, self._client, job_id = self._job_id, disallow_discard = self._disallow_discard)
						self._result = [0, 'restore_resume']
				
				else:
					self._restore_engine.resume_restore(client = self._client, disallow_discard = self._disallow_discard)
					self._result = [0, 'restore_resume']
					
			except Exception as ex:
				logger.error('Couldn\'t initialize restore on server side')
				self._result = [4, ex.message]
				return
				
			try:
				self._restore_engine.run()
				self._client.disconnect()
				
			except Exception as ex:

				self._index = self._restore_engine.get_index_new()
				logger.error('Restore of {} unexpectly end ({}/{})'.format(self._name, *self._restore_engine.get_restore_status()))
				self._result = [5, ex.message]
				return
				
			restore_result = self._restore_engine.get_result()
			if(restore_result[0] == 4):
				self._index = self._restore_engine.get_index_new()
				logger.error('Restore of {} end ({}/{}) backup image broken'.format(self._name, *self._restore_engine.get_restore_status()))
				self._result = [2, 'index failed']
				return
				
			if(restore_result[0] == 3):
				self._index = self._restore_engine.get_index_new()
				logger.error('Restore of {} unexpectly end ({}/{})'.format(self._name, *self._restore_engine.get_restore_status()))
				self._result = [5, 'io_error']
				return
				
			elif(restore_result[0] == 2):
				self._index = self._restore_engine.get_index_new()
				logger.error('Restore of {} end, worker lost connection ({}/{})'.format(self._name, *self._restore_engine.get_restore_status()))
				self._result = [5, 'unexpectly_end']
				return
				
			elif(restore_result[0] == 1):
				self._index = self._restore_engine.get_index_new()
				logger.warning('Restore of {} breake ({}/{})'.format(self._name, *self._restore_engine.get_restore_status()))
				self._result = [5, 'breake']
				return
				
			self._result = [0, 'restore_complete']
			
			logger.info('Finished restore of {}'.format(self._name))
			self._result = [0, 'completed']
			
				
		except Exception as ex:
			logger.error('Restore end on uncaught exception: {}'.format(ex.message))
			self._result = [1, ex.message]

class BackupJob(threading.Thread):
	"""docstring for BackupJob"""
	def __init__(self, name, btype = 'auto', flags = 0, metadata = {}, snapshot = None, prev_index = {}, backup_identifier = None, origin = None, backup_engine = None, job_id = None, uuid = None):
		assert type(name) == str, "Invalid name type"
		if(job_id == None):
			self._job_id = os.urandom(8).encode('hex')
			
		else:
			self._job_id = job_id
			
		super(BackupJob, self).__init__(name='backup_thread_{}_{}'.format(name,self._job_id))
		if(btype == 'image'):
			self._name = 'image'
			self._orig_name = name
		else:
			self._name = name
			self._orig_name = None
			
		self._type = btype
		self._flags = flags
		self._metadata = metadata
		
		self._index = prev_index
		
		self._client = None
		self._backup_engine = backup_engine
		self._snapshot = snapshot
		self._backup_identifier = backup_identifier
		self._origin = origin
		self._cancel = False
		self._uuid = uuid
		
		self._result = [0, 'init']
		# result - [error_code, text_status]
		# 0, status - OK
		# 1, exception - Uncaught exception
		# 2, reason - Backup couldn't start
		# 3, reason - Couldn't connect to server
		# 4, reason - Error during backup init on server side
		# 5, reason - Backup brake during process save data to continue
		# 6, reason - Error occured during finalize backup
		
	def __reduce__(self):
		return (self.__class__, (self.get_name(), self._type, self._flags, self._metadata, self._snapshot, self._index, self._backup_identifier, self._origin, None, self._job_id, self._uuid))
	
	def get_copy(self):
		retVal = self.__class__(self.get_name(), self._type, self._flags, self._metadata, self._snapshot, self._index, self._backup_identifier, self._origin, self._backup_engine, self._job_id, self._uuid)
		return retVal
		
	def stop(self):
		self._backup_engine.stop()
		
	def cancel(self):
		self._backup_engine.cancel()
		self._cancel = True
		
	def get_backup_identifier(self):
		return self._backup_identifier
		
	def get_snapshot(self):
		return self._snapshot
		
	def run(self):
		try:
			if(len(self._index) == 0):
				logger.info('Started backup of {}'.format(self._name))
				
			else:
				logger.info('Started backup of {} from block {}'.format(self._name, len(self._index)))
			
			try:
				if(self._type == 'image'):
					domain_uuid = virt.guest.get_uuid_from_name(self._orig_name)
				
				else:	
					domain_uuid = virt.guest.get_uuid_from_name(self._name)
					
			except:
				domain_uuid = None
				
			if domain_uuid != None:
				with c2padmin.config.node_config_lock:
					c2padmin.config.node_config['libvirt']['domains'][domain_uuid]['backups']['running'].append(self)
					
			else:
				logger.error('Domain {} not found'.format(self._name))
				self._result = [2, 'domain_not_found']
				return
			
			if(not (self._flags & FLAG_NO_SNAPSHOT)):				
				if(self._snapshot == None):
					self._snapshot = virt.volume.create_snapshot(self._name)
					
				self._result = [0, 'snapshot_created']	
				source_file = os.path.join('/dev/HostVG',self._snapshot)
			else:
				source_file = os.path.join('/dev/HostVG',self._name)
				
			try:
				assert os.path.exists(source_file), "Source file does not exist"
				assert not os.path.isdir(source_file), "Source file is directory"
			
			except Exception as ex:
				logger.error('Source device {} not found'.format(source_file))
				self._result = [2, 'source_not_found']
				return
				
			self._result = [0, 'prepared']
			
			self._client = client.BackupClient()
			
			try:
				self._client.connect()
				self._result = [0, 'connected']
				
			except Exception as ex:
				logger.error('Couldn\'t connect to backup server')
				self._result = [3, ex.message]
				return
			
			try:
				if(self._backup_engine == None):	
					self._backup_engine = backup.Backup_Engine(self._client)
					if(len(self._index) == 0):
						self._backup_identifier, self._origin = self._backup_engine.start_backup(self._name, source_file, self._type, orig_name = self._orig_name, metadata = self._metadata, job_id = self._job_id)
						self._result = [0, 'backup_start']
						
					else:
						self._backup_identifier, self._origin = self._backup_engine.resume_backup(self._name, source_file, self._type, self._backup_identifier, self._origin, self._index, self._job_id)
						self._result = [0, 'backup_resume']
				
				else:
					self._backup_identifier, self._origin = self._backup_engine.resume_backup(client = self._client)
					self._result = [0, 'backup_resume']
					
			except Exception as ex:
				logger.error('Couldn\'t initialize backup on server side')
				self._result = [4, ex.message]
				return
			
			try:
				self._backup_engine.run()
				self._client.disconnect()
				
			except Exception as ex:
				self._index = self._backup_engine.get_index_new()
				logger.error('Backup of {} unexpectly end ({}/{})'.format(self._name, *self._backup_engine.get_write_status()))
				self._result = [5, ex.message]
				return
			
			backup_result = self._backup_engine.get_result()
			if(backup_result[0] == 3):
				self._index = self._backup_engine.get_index_new()
				logger.error('Backup of {} unexpectly end ({}/{})'.format(self._name, *self._backup_engine.get_write_status()))
				self._result = [5, 'unexpectly_end']
				return
				
			elif(backup_result[0] == 2):
				self._index = self._backup_engine.get_index_new()
				logger.error('Backup of {} end, worker lost connection ({}/{})'.format(self._name, *self._backup_engine.get_write_status()))
				self._result = [5, 'unexpectly_end']
				return
				
			elif(backup_result[0] == 1):
				self._index = self._backup_engine.get_index_new()
				logger.warning('Backup of {} breake ({}/{})'.format(self._name, *self._backup_engine.get_write_status()))
				self._result = [5, 'breake']
				return
				
			self._result = [0, 'backup_complete']
			
			if self._snapshot != None:
				try:
					virt.volume.remove_volume(self._snapshot)
					self._snapshot = None
					
				except Exception as ex:
					logger.warning('Couldn\'t remove snapshot {}'.format(self._snapshot))
					self._result = [6, ex.message] 
				
			try:
				c2padmin.config.node_config['libvirt']['domains'][domain_uuid]['backups']['completed'].refresh()
				with c2padmin.config.node_config_lock:
					try:
						original_size = len(c2padmin.config.node_config['libvirt']['domains'][domain_uuid]['backups']['running'])
						c2padmin.config.node_config['libvirt']['domains'][domain_uuid]['backups']['running'] = filter(lambda x: x.get_job_id() != self._job_id, c2padmin.config.node_config['libvirt']['domains'][domain_uuid]['backups']['running'])
						
						if(original_size == len(c2padmin.config.node_config['libvirt']['domains'][domain_uuid]['backups']['running'])):
							raise Exception('Couldn\'t remove job form running')
						
					except Exception as ex:
						logger.warning('Couldn\'t remove job form running')
						if(self._result[0] == 6):
							if(type(self._result[1]) != list):
								self._result[1] = [self._result[1]]
								
							self._result[1].append(ex.message)
							
						else:	
							self._result = [6, ex.message]

			except Exception as ex:
				logger.warning('Finalize failed')
				if(self._result[0] == 6):
					if(type(self._result[1]) != list):
						self._result[1] = [self._result[1]]
						
					self._result[1].append(ex.message)
					
				else:	
					self._result = [6, ex.message]
					
				print self._result
					
			
			if(self._result[0] == 6):
				logger.warning('Backub of {} finished with error during finalize (node_config could be broken)'.format(self._name))
				
			else:
				if(self._cancel):
					logger.info('Backup of {} cancel'.format(self._name))
					self._result = [0, 'cancel']
					
				else:	
					logger.info('Finished backup of {}'.format(self._name))
					self._result = [0, 'completed']
			
		except Exception as ex:
			logger.error('Backup end on uncaught exception: {}'.format(str(ex)))
			if self._snapshot != None:
				try:
					virt.volume.remove_volume(self._snapshot)
					self._snapshot = None
					
				except Exception as ex1:
					logger.warning('Couldn\'t remove snapshot {} ({})'.format(self._snapshot, ex1.message))
			
			
			self._result = [1, ex.message]

	def get_name(self):
		if(self._orig_name == None):
			return self._name
			
		else:
			return self._orig_name
		
	def get_job_id(self):
		return self._job_id
		
	def get_uuid(self):
		return self._uuid
		
	def get_result(self):
		return self._result
		
	def get_write_status(self):
		return self._backup_engine.get_write_status() if self._backup_engine != None else [None, None]
		
	def get_dict(self):
		return {'__BackupJob__':True,'name':self._name,'thread_id':self._job_id, 'type': self._type, 'flags': self._flags, 'metadata': self._metadata, 'snapshot': self._snapshot}

class BackupSchedule(object):
	"""docstring for BackupSchedule"""
	def __init__(self, backup_list, domain_name):
		super(BackupSchedule, self).__init__()
		self._backup_list = backup_list
		self._domain_name = domain_name
		self._parser = ConfigParser.ConfigParser(allow_no_value=True)
		self.load_config()
		#if init_dict != None:
		#	self.set_dict(init_dict)

	def load_config(self):
		self._parser.read(os.path.join(os.path.dirname(__file__),'conf/backup_defaults.ini'))
		config_file = '/etc/c2padmin/backups/' + self._domain_name + '.ini'
		self._parser.read(config_file)
		self._enabled = self._parser.getboolean('scheduler', 'enabled')
		self._interval = self._parser.getint('scheduler', 'interval')
		self._keep_rules = map(lambda x: tuple(map(lambda x: int(x.strip()), x.split(':'))), self._parser.get('scheduler', 'keep_rules').split(','))
		self._time_of_day = map(lambda x: tuple(map(lambda y: datetime.time(*map(int, y.strip().split(':'))), x.split('-'))), self._parser.get('scheduler', 'time_of_day').split(','))
		self._day_of_week = map(int, self._parser.get('scheduler', 'day_of_week').split(','))
		self.save_config()
	
	def save_config(self):
		config_file = '/etc/c2padmin/backups/' + self._domain_name + '.ini'
		config_file_fh = open(config_file, 'w')
		self._parser.write(config_file_fh)
		config_file_fh.close()
		
	def get_dict(self):
		"""docstring for get_dict"""
		return {
			'__BackupSchedule__': True,
			'enabled': self._enabled,
			'interval': self._interval,
			'keep_rules': self._keep_rules,
			'time_of_day': self._time_of_day,
			'day_of_week': self._day_of_week,
		}
		
	def set_dict(self, init_dict):
		assert type(init_dict) == dict, 'Invalid `init_dict` type'
		assert 'enabled' not in init_dict or type(init_dict['enabled']) == bool, 'Invalid `enabled` type'
		assert 'interval' not in init_dict or type(init_dict['interval']) == int, 'Invalid `interval` type'
		assert 'keep_rules' not in init_dict or (type(init_dict['keep_rules']) == list and len(init_dict['keep_rules']) >= 1 and len(init_dict['keep_rules']) == len(filter(lambda x: type(x) == list and len(x) == 2 and type(x[0]) == int and type(x[1]) == int, init_dict['keep_rules']))), 'Invalid `keep_rules` type'
		assert 'time_of_day' not in init_dict or (type(init_dict['time_of_day']) == list and len(init_dict['time_of_day']) >= 1 and len(init_dict['time_of_day']) == len(filter(lambda x: type(x) == list and len(x) == 2 and type(x[0]) in [str, unicode] and type(x[1]) in [str, unicode], init_dict['time_of_day']))), 'Invalid `time_of_day` type'
		assert 'day_of_week' not in init_dict or (type(init_dict['day_of_week']) == list and len(init_dict['day_of_week']) == len(set(init_dict['day_of_week'])) == len(filter(lambda x: type(x) == int and x >= 0 and x <= 6, init_dict['day_of_week']))), 'Invalid `day_of_week` type'
	
		if('enabled' in init_dict):
			self._enabled = init_dict['enabled']
			self._parser.set('scheduler', 'enabled', self._enabled)
			
		if('interval' in init_dict):
			self._interval = init_dict['interval']
			self._parser.set('scheduler', 'interval', self._interval)
			
		if('keep_rules' in init_dict):
			self._keep_rules = init_dict['keep_rules']
			self._parser.set('scheduler', 'keep_rules', ','.join(map(lambda x: str(x[0]) + ':' + str(x[1]), self._keep_rules)))
			
		if('time_of_day' in init_dict):
			init_dict['time_of_day'] = map(lambda x: [datetime.time(*map(int, x[0].split(':'))), datetime.time(*map(int, x[1].split(':')))], init_dict['time_of_day'])
			self._time_of_day = init_dict['time_of_day']
			self._parser.set('scheduler', 'time_of_day', ','.join(map(lambda x: x[0].strftime('%H:%M') + '-' + x[1].strftime('%H:%M'), self._time_of_day)))
			
		if('day_of_week' in init_dict):
			self._day_of_week = init_dict['day_of_week']
			self._parser.set('scheduler', 'day_of_week', ','.join(map(str, init_dict['day_of_week'])))
			
		# TODO ulozit nastaveny data
		
		self.save_config()
		return True

	def backup_now(self):
		"""docstring for backup_now"""
		if not self._enabled:
			return False
		
		if(self._backup_list == None or not self._backup_list.is_available()):
			return False
		
		now = datetime.datetime.now()
		found_time = False
		for time_interval in self._time_of_day:
			if now.time() >= time_interval[0] and now.time() <= time_interval[1]:
				found_time = True
				break
		
		if not found_time:
			logger.debug('Will not backup at this time of day')
			return False
		
		if now.weekday() not in self._day_of_week:
			logger.debug('Will not backup on this weekday')
			return False
		
		last_backup = self._backup_list.get_last_backup('auto')
		if last_backup != None:
			if last_backup['time'] > (time.time()-self._interval):
				logger.debug('Latest backup is fresh enough. Will not backup now')
				return False
		return True
	
	def delete_old(self):
		"""docstring for delete_old"""
		# TODO - tohle udelat vyrazne chytrejsi
		if(self._backup_list == None or not self._backup_list.is_available()):
			return
			
		try:
			_client = client.BackupClient()
			_client.connect()
			
		except:
			return
			
		num_periods = len(self._keep_rules)
		backups_by_period = [[]]*num_periods
		mytime = time.time()
		for backup in self._backup_list.get_list():
			if backup['type'] != 'auto':
				continue
				
			for i in range(num_periods):
				if i == num_periods-1 or (mytime - backup['time']) <= (self._keep_rules[i][0]*self._keep_rules[i][1]):
					backups_by_period[i].append(backup)
					break
		
		for i in range(num_periods):
			while len(backups_by_period[i]) > self._keep_rules[i][1]:
				backup_to_delete = backups_by_period[i].pop(0)
				logger.info('Deleting backup {} {}'.format(backup_to_delete['name'],backup_to_delete['path']))
				_client.removeBackup(backup_to_delete['name'],backup_to_delete['path'])
				
		self._backup_list.refresh()
		
