import OpenSSL,os
from Crypto.Util import asn1
import config,logger

logger = logger.getLogger(__name__)

def read_cert(filename):
	"""docstring for read_cert"""
	try:
		fd = open(filename)
		buff = fd.read()
		fd.close()
	
		cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM,open(filename).read())
		return cert
	except:
		return None

def new_request(feature,key_size=2048):
	key = OpenSSL.crypto.PKey()
	key.generate_key(OpenSSL.crypto.TYPE_RSA,key_size)
	req = OpenSSL.crypto.X509Req()
	req.set_pubkey(key)
	subject = req.get_subject()
	subject.commonName = config.node_config['uuid']
	req.sign(key,'sha256')
	# save key
	key_file = open(config.config.get_certificate_storage_path(feature)+'.key','w')
	key_file.write(OpenSSL.crypto.dump_privatekey(OpenSSL.crypto.FILETYPE_PEM,key))
	key_file.close()
	os.chmod(config.config.get_certificate_storage_path(feature)+'.key',0600)
	return dump_certificate_request(req)

def dump_certificate_request(req):
	"""docstring for dump_certificate_request"""
	return OpenSSL.crypto.dump_certificate_request(OpenSSL.crypto.FILETYPE_PEM,req)
	
def load_certificate(buff):
	"""docstring for load_certificate"""
	return OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, buff)
	
def save_certificate(cert,feature):
	"""docstring for save_certificate"""
	filename_base = config.config.get_certificate_storage_path(feature)
	if type(cert) != OpenSSL.crypto.X509:
		cert = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM,cert)
	if not os.path.exists(filename_base+'.key'):
		logger.debug('No private key')
		return False
	priv_key = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM,open(filename_base+'.key').read())
	if not check_keys(cert,priv_key):
		logger.debug('Keys don\'t match')
		return False
	cert_file = open(filename_base+'.crt','w')	
	cert_file.write(OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_PEM,cert))
	cert_file.close()
	return None
	
def check_keys(cert, priv):
	"""docstring for check_keys"""

	c=OpenSSL.crypto

	pub=cert.get_pubkey()

	# Only works for RSA (I think)
	if pub.type()!=c.TYPE_RSA or priv.type()!=c.TYPE_RSA:
	    raise Exception('Can only handle RSA keys')

	# This seems to work with public as well
	pub_asn1=c.dump_privatekey(c.FILETYPE_ASN1, pub)
	priv_asn1=c.dump_privatekey(c.FILETYPE_ASN1, priv)

	# Decode DER
	pub_der=asn1.DerSequence()
	pub_der.decode(pub_asn1)
	priv_der=asn1.DerSequence()
	priv_der.decode(priv_asn1)

	# Get the modulus
	pub_modulus=pub_der[1]
	priv_modulus=priv_der[1]

	if pub_modulus==priv_modulus:
	    return True
	else:
	    return False