import Queue,threading,traceback
import virt, config, util, time
import imp, importlib

from c2padmin import logger as logmodule

logger = logmodule.getLogger(__name__)

import c2padmin

def fetch_completed_commands():
	"""docstring for fetch_complete_commands"""
	with config.node_config_lock:
		commands = []
		for i in range(config.cc.completed_size()):
			try:
				commands.append(config.cc.get(timeout=0.1))
				
			except Queue.Empty:
				break
				
		config.node_config['commands_in_queue'] = False
		return commands

class VPSBusy(Exception):
	def __init__(self, timeout = 30):
		self._timeout = timeout
		
	def timeout(self):
		return self._timeout

class CommandProgress(object):
	"""docstring for CommandProgress"""
	def __init__(self, command_id=None):
		super(CommandProgress, self).__init__()
		self._command_id = command_id
		self._progress = None
	
	def set(self,progress):
		"""docstring for set"""
		self._progress = progress
		
	def get(self):
		"""docstring for get"""
		return self._progress
	
	def __str__(self):
		return str(self._progress)


class CommandsCaller(object):
	def __init__(self, modules = ['virt','virt.guest', 'util']):
		self._modules = modules
		self._queues = {}
		self._workers = {}
		self._completed = Queue.Queue()
		self._main_lock = threading.Lock()

		self._progress_data = {}
		self._in_processing = []
		self._lock = {}
		with config.node_config_lock:
			config.node_config['commands_progress'] = self._progress_data
			

	def put(self, job, block = True, timeout = None):
		assert 'recid' in job, 'recid not found in job'
		assert 'function_name' in job, 'function_name not found in job'
		assert 'params' in job, 'params not found in job'
		if('id_vserver_c2p' in job):
			worker_id = job['id_vserver_c2p']

		else:
			worker_id = None

		with self._main_lock:
			for tmp_id in self._workers.keys():
				if(not self._workers[tmp_id].is_alive()):
					del self._workers[tmp_id]
					if(tmp_id != worker_id and tmp_id in self._queues and self._queues[tmp_id].empty()):
						del self._queues[tmp_id]

			if job['recid'] in self._in_processing:
				return None

			if(worker_id not in self._queues):
				self._queues[worker_id] = Queue.Queue()

			self._queues[worker_id].put(job, block, timeout)
			self._in_processing.append(job['recid'])

			if(worker_id not in self._workers):
				name = 'commands_worker' + ('_' + str(worker_id) if(worker_id != None) else '')
				self._workers[worker_id] = threading.Thread(name = name, target = self._run, args = (worker_id,))
				self._workers[worker_id].start()

		return None
		
	def put_complete(self, job):
		self._completed.put(job)
		return True

	def get(self, block=True, timeout=None):
		job = self._completed.get(block,timeout)

		with self._main_lock:
			try:
				self._in_processing.remove(job['recid'])
				
			except:
				pass
				
			if self._progress_data.has_key(job['recid']):
				del self._progress_data[job['recid']]

		return job

	def completed_size(self):
		return self._completed.qsize()
		
	def get_worker_lock(self, worker_id):
		with self._main_lock:
			if(worker_id not in self._lock):
				self._lock[worker_id] = threading.Lock()
				
		return self._lock[worker_id]

	def _run(self, worker_id):
		logger.debug('Starting commands caller worker {}'.format(threading.current_thread().getName()))
		job = None
		try:
			while True:
				if config.main_thread_signal != None:
					# TODO
					return

				if(worker_id in self._lock):
					self._lock[worker_id].acquire()
					self._lock[worker_id].release()

				try:
					self._main_lock.acquire()
					self._main_lock.release()
				
					if(job == None):
						job = self._queues[worker_id].get(True, 10)

				except Queue.Empty:
					break

				try:
					ret = self.call(job['recid'],job['function_name'],job['params'])
				
				except VPSBusy as ex:
					logger.info('Catch VPSBusy exception. Sleep for {}s'.format(ex.timeout()))
					time.sleep(ex.timeout())
					continue

				except Exception as ex:
					ret = ['exception', str(ex), traceback.format_exc()]

				job['returned'] = ret
				print ret
				with config.node_config_lock:
	 				config.node_config['commands_in_queue'] = True

				self._completed.put(job)
				job = None
				logger.info('Job complete (SYNC)')
				#config.config.update_node_status()
				
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
			
		logger.debug('Ending commands caller worker {}'.format(threading.current_thread().getName()))

	def call(self, recid, func_name, args):
		func_name = func_name.rsplit('.', 1)
		if(len(func_name) != 2):
			raise Exception('Invalid function call: function unknown')

		module_name = func_name[0]
		func_name = func_name[1]

		if module_name not in self._modules:
			raise Exception('Invalid function call: permission denied')

		module = importlib.import_module('c2padmin.' + module_name)

		func = getattr(module,func_name)
		if(func == None):
			raise Exception('Invalid function call: function unknown')

		progress = CommandProgress(recid)
		with config.node_config_lock:
			self._progress_data[recid] = progress

		if(type(args) == dict):
			if('progress' in args):
				raise Exception('Args name progress is reserved')

			args['progress'] = progress
			return func(**args)

		else:
			return func(*args,progress=progress)
			