import threading,hashlib,cPickle,time,ConfigParser,uuid,subprocess,os,re,logging,signal,cache,copy,socket
try:
	import libvirt
except:
	pass
import util,cert,virt
from backups.types import BackupSchedule,BackupList
import backups.client
import logger as logmodule
from xml.dom.minidom import parseString
from version import __version__

import c2padmin

logger = logmodule.getLogger(__name__)

API_KEY_LENGTH = 32
LVS_LV_NAME = 0
LVS_VG_NAME = 1
LVS_ATTR = 2
LVS_SIZE = 3
PVS_PV_NAME = 0
PVS_VG_NAME = 1
PVS_FORMAT = 2
PVS_ATTR = 3
PVS_SIZE = 4
PVS_FREE = 5
VGS_VG_NAME = 0
VGS_PV_COUNT = 1
VGS_LV_COUNT = 2
VGS_SN_COUNT = 3
VGS_ATTR = 4
VGS_SIZE = 5
VGS_FREE = 6

NODE_FLAGS_CONFIG_UPDATED = 1
NODE_FLAGS_MASTER_CANDIDATE = {'api':2,'vpn':4,'dummy0':8,'dummy1':16,'dummy2':32,'dummy3':64,'dummy4':128,'dummy5':256}
NODE_FLAGS_TERMINATING = 512

MEMINFO_KEYS = ['MemTotal','SwapTotal','MemFree']
CPUINFO_BLACKLIST_KEYS = ['cpu MHz']

ST_STARTING = 0
ST_UP = 1
ST_DOWN = 2
ST_REINSTALL = 3

node_uuid_bytes = None
api_key = None
node_status = {
	'flags': 0,
	'node_config_hash': None,
	'disconnect_node_data': {},
	'sync_network_timestamp': 0,
	'master_call_timestamp': 0,
	'active_snapshots': {}
}
node_status_lock = util.LoggingLock('status_lock')
node_config = {
	'preinstall':False,
	'lvm':{
		'lvs':None,
		'vgs':None,
		'pvs':None
	},
	'roles':{
		'api':None,
		'vpn':None
	},
	'roles_prio':{
		
	},
	'certificates':{
		'backup':None,
		'vpn':None,
		'monitoring':None
	},
	'ssh_keys': {
		'pub': None,
		'priv': None
	},
	'info': {
		'version': __version__,
		'state': 0,
		'timestamp': time.time()
	},
	'net_identifier':None,
	'hostname': socket.gethostname(),
	'uuid':None,
	'commands_in_queue':False,
	'bridge_config':{'serial':-1,'data':None},
	'peer_ips':None,
	'libvirt':{'libvirt_version':None,'domains':{}},
	'backup_server': {'unavailable_timestamp': None},
	'remote_shell': [],
	'active_snapshots': {},
	'uptime': 0,
	'cpu_usage': {'usage': 0, 'usage_percent': 0, 'cores': []},
	'dhcp': {
		'hash': None,
		'reload_config': False,
		'config': {
			'subnets': {},
			'hosts': {},
			'vxlan_id': None
		}
	}
}
node_config_lock = util.LoggingLock('node_config')
node_system_locks = {
	'sqlite': threading.Lock(),
	'xmlrpc': {},
	'clone_migrate': threading.Lock()
}
master_candidature_time = {'api':None,'vpn':None}
main_thread_signal = None
master_status_lock = util.LockNB()

def get_network_identifier(myuuid):
	"""get 24bit network identifier from uuid"""
	if type(myuuid) == type(''):
		myuuid = uuid.UUID(myuuid)
	return myuuid.fields[5]>>24

class Config():
	"""docstring for Config"""
	def __init__(self,filename=None):
		self.parser = None
		self.parse_ini(filename)
		self.realod_backup_lists = False
		tmp = cache.cache.load_config('node_config')
		if(tmp == False):
			logger.error('Unable to parse previous node_config')
			
		else:
			if(tmp != None):
				with node_config_lock:
					node_config = tmp
				
				cache.cache.commit_config('node_config')
					
		tmp = cache.cache.load_config('node_status')
		if(tmp == False):
			logger.error('Unable to parse previous node_status')
			
		else:
			if(tmp != None):
				with node_status_lock:
					node_status = tmp
				
				cache.cache.commit_config('node_status')
				
	
	def status_loop(self):
		"""docstring for status_loop"""
		try:
			while True:
				if main_thread_signal != None:
					if(main_thread_signal == signal.SIGHUP):
						with node_config_lock:
							cache.cache.save_config('node_config', node_config)
						
						with node_status_lock:
							cache.cache.save_config('node_status', node_status)
				
					return
					
				logger.debug('Running node status')
				self.get_node_status()
				#self.update_node_status()
				time.sleep(5)
		
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
	
	def sync_config_loop(self):
		try:
			while True:
				if main_thread_signal != None:
					return
				
				self.sync_config()	
				time.sleep(2)
		
		except Exception as ex:	
			c2padmin.dispatcher.unexpectly_end(ex)
			
	def load_ssh_keys(self):
		pub = None
		priv = None
		try:
			fh = open('/root/.ssh/id_rsa', 'r')
			priv = fh.read()
			fh.close()
			
		except:
			with node_config_lock:
				node_config['ssh_keys'] = {
					'pub': None,
					'priv': None
				}
				
			return
				
		try:
			fh = open('/root/.ssh/authorized_keys', 'r')
			lines = fh.readlines()
			fh.close()
			for line in lines:
				line = line.strip()
				tmp = line.split(None, 3)
				if(len(tmp) == 3 and tmp[0] == 'ssh-rsa' and tmp[2] == 'locality_key'):
					pub = line
			
		except:
			with node_config_lock:
				node_config['ssh_keys'] = {
					'pub': None,
					'priv': None
				}
			
			return
				
		if(pub == None or priv == None):
			with node_config_lock:
				node_config['ssh_keys'] = {
					'pub': None,
					'priv': None
				}
				
			return
				
		with node_config_lock:
			node_config['ssh_keys'] = {
				'pub': pub,
				'priv': priv
			}
			
	def save_ssh_keys(self):
		pub = None
		priv = None
		with node_config_lock:
			pub = node_config['ssh_keys']['pub']
			priv = node_config['ssh_keys']['priv']
		
		if(pub == None or priv == None):
			return False
			
		try:
			fh = open('/root/.ssh/id_rsa', 'w')
			fh.write(priv)
			fh.close()
			os.chmod('/root/.ssh/id_rsa', 0600)
			
		except:
			return False
		
		lines = []	
		try:
			fh = open('/root/.ssh/authorized_keys', 'r')
			lines = fh.readlines()
			fh.close()
			
		except:
			pass
			
		try:
			fh = open('/root/.ssh/authorized_keys', 'w')
			done = False
			for line in lines:
				line = line.strip()
				tmp = line.split(None, 3)
				if(len(tmp) == 3 and tmp[0] == 'ssh-rsa' and tmp[2] == 'locality_key'):
					fh.write(pub + '\n')
					done = True
					
				else:
					fh.write(line + '\n')
			
			if(not done):
				fh.write(pub + '\n')
			
			fh.close()
			os.chmod('/root/.ssh/authorized_keys', 0600)
			
		except:
			raise
			print 'authorized_keys'
			return False
			
		return True
			
	def sync_config(self):
		node_config_hash = hashlib.sha1()
		node_config_hash.update(cPickle.dumps(node_config))

		send = False
		if(node_status_lock.acquire(False)):
			if(node_status['node_config_hash'] != node_config_hash.hexdigest()):
				node_status['flags'] = node_status['flags'] | NODE_FLAGS_CONFIG_UPDATED
				node_status['node_config_hash'] = node_config_hash.hexdigest()
				send = True
		
			if(send):
				logger.debug('Data changed. Config %s send now.' % (str(node_config_hash.hexdigest()),))
				try:
					util.keepalive.send_now()

				except:
					pass
			
			node_status_lock.release()
						
		del node_config_hash	
	
	def get(self,section,key,itype=str):
		"""get config value read from ini"""
		assert type(itype) == type, 'Invalid data type requested %s' % (itype,)
		if(itype == bool):
			return self.parser.getboolean(section,key)
			
		return itype(self.parser.get(section,key))
	
	def parse_ini(self,filename=None):
		global node_uuid_bytes,api_key
		"""parse ini files, if you want non-standard ini file to be loded pass it in 'filename' argument"""
		parser = ConfigParser.ConfigParser(allow_no_value=True)
		parser.readfp(open(os.path.join(os.path.dirname(__file__),'conf/defaults.ini')))
		config_files = []
		
		config_files = ['/etc/c2padmin/node.ini']
		
		if filename != None:
			config_files.append(filename)
		
		if len(config_files) > 0:
			parser.read(config_files)
		try:
			log_level = parser.get('daemon','log_level')
			logmodule.logger_root.setLevel(getattr(logging,log_level.upper()))
		except:
			parser.set('daemon','log_level','debug')
		try:
			myuuid = uuid.UUID(parser.get('node','uuid'))
			parser.set('network','identifier',str(get_network_identifier(myuuid)))
		except:
			parser.set('node','uuid',str(uuid.uuid1((uuid.getnode() & 0xffffff) | (int(parser.get('network','identifier')) << 24))))
		
		# certificate locations
		if not parser.has_option('node','ca_bundle'):
			parser.set('node','ca_bundle',os.path.join(parser.get('node','cert_path'),'ca_bundle.crt'))
		if not parser.has_option('node','backup_cert'):
			parser.set('node','backup_cert',os.path.join(parser.get('node','cert_path'),'backup.crt'))
		if not parser.has_option('node','backup_key'):
			parser.set('node','backup_key',os.path.join(parser.get('node','cert_path'),'backup.key'))
		if not parser.has_option('node','vpn_cert'):
			parser.set('node','vpn_cert',os.path.join(parser.get('node','cert_path'),'vpn.crt'))
		if not parser.has_option('node','vpn_key'):
			parser.set('node','vpn_key',os.path.join(parser.get('node','cert_path'),'vpn.key'))
			
		# public backup
		if(not parser.has_option('node','backup_public_cert')):
			parser.set('node','backup_public_cert', '')
			
		if(not parser.has_option('node','backup_public_cert')):
			parser.set('node','backup_public_key', '')
			
		if(not parser.has_option('node','backup_public_password')):
			parser.set('node','backup_public_password', '')
		
		self.parser = parser
		
		uu = uuid.UUID(self.get('node','uuid'))
		with node_status_lock:
			node_status['uuid'] = str(uu)
			
		with node_config_lock:
			if uu.version != 1:
				node_config['preinstall'] = True
				node_config['roles']['api'] = 'master'
				node_config['roles']['vpn'] = 'disabled'
			else:
				node_config['roles']['api'] = self.get('daemon','api_role')
				if self.get('network','public',int):
					node_config['roles']['vpn'] = 'disabled'
				else:
					node_config['roles']['vpn'] = self.get('daemon','vpn_role')
			node_config['uuid'] = node_status['uuid']
			node_config['net_identifier'] = int(self.get('network','identifier'))
		self.reload_certificates(run_node_status=False)
		self.load_ssh_keys()
		
		node_uuid_bytes =uu.bytes
		try:
			k = self.get('node','api_key').decode('hex')
			if len(k) != API_KEY_LENGTH:
				raise Exception('Invalid key length')
			api_key = k
		except:
			pass
	
	def get_carbon_destinations(self, conf_dir = '/opt/graphite/conf/'):
		try:
			relay_rules_c = ConfigParser.ConfigParser(allow_no_value=True)
			relay_rules_c.readfp(open(conf_dir + '/relay-rules.conf'))
			destinations = relay_rules_c.get('default', 'destinations')
			destinations = map(lambda x: x.strip(), destinations.split(','))
			return destinations
			
		except:
			return False
			
	def set_carbon_destinations(self, dest_list, conf_dir = '/opt/graphite/conf/'):
		try:
			relay_rules_c = ConfigParser.ConfigParser(allow_no_value=True)
			relay_rules_c.readfp(open(conf_dir + '/relay-rules.conf'))
			relay_rules_c.set('default', 'destinations', ', '.join(dest_list))
			relay_rules_c.write(open(conf_dir + '/relay-rules.conf', 'w'))
			
			carbon_c = ConfigParser.ConfigParser(allow_no_value=True)
			carbon_c.readfp(open(conf_dir + '/carbon.conf'))
			carbon_c.set('relay', 'DESTINATIONS', ', '.join(dest_list))
			carbon_c.set('relay', 'REPLICATION_FACTOR', len(dest_list))
			carbon_c.write(open(conf_dir + '/carbon.conf', 'w'))
			
			if(os.system('systemctl restart carbon-relay') != 0):
				logger.error('carbon-relay restart failed')
				return False
			
			logger.info('carbon-relay new destinations: ' + ', '.join(dest_list))
			
			return True
			
		except:
			logger.error('carbon-relay set  new destinations failed')
			return False
	
	def reload_certificates(self,run_node_status=True):
		"""docstring for reload_certificates"""
		backup_cert = None
		if os.path.exists(self.get_certificate_storage_path('backup.crt')) and os.path.exists(self.get_certificate_storage_path('backup.key')):
			try:
				backup_crt = cert.read_cert(self.get_certificate_storage_path('backup.crt'))
				if backup_crt.get_subject().CN != node_config['uuid']:
					raise Exception('Invalid commonName')
				if backup_crt.has_expired():
					raise Exception('Certificate expired')
				
				backup_cert = {
					'filename': self.get_certificate_storage_path('backup.crt'),
					'digest': backup_crt.digest('sha256'),
					'serial': backup_crt.get_serial_number(),
					'valid':(backup_crt.get_notBefore(),backup_crt.get_notAfter()),
				}
				
			except:
				backup_cert = None
		
		vpn_cert = None		
		if os.path.exists(self.get_certificate_storage_path('vpn.crt')) and os.path.exists(self.get_certificate_storage_path('vpn.key')):
			try:
				vpn_crt = cert.read_cert(self.get_certificate_storage_path('vpn.crt'))
				if vpn_crt.get_subject().CN != node_config['uuid']:
					raise Exception('Invalid commonName')
				if vpn_crt.has_expired():
					raise Exception('Certificate expired')
				
				vpn_cert = {
					'filename': self.get_certificate_storage_path('vpn.crt'),
					'digest': vpn_crt.digest('sha256'),
					'serial': vpn_crt.get_serial_number(),
					'valid':(vpn_crt.get_notBefore(),vpn_crt.get_notAfter()),
				}
				
			except:
				vpn_cert = None
		
		monitoring_cert = None		
		if os.path.exists(self.get_certificate_storage_path('monitoring.crt')) and os.path.exists(self.get_certificate_storage_path('monitoring.key')):
			try:
				monitoring_crt = cert.read_cert(self.get_certificate_storage_path('monitoring.crt'))
				if monitoring_crt.get_subject().CN != node_config['uuid']:
					raise Exception('Invalid commonName')
				if monitoring_crt.has_expired():
					raise Exception('Certificate expired')
				
				monitoring_cert = {
					'filename': self.get_certificate_storage_path('monitoring.crt'),
					'digest': monitoring_crt.digest('sha256'),
					'serial': monitoring_crt.get_serial_number(),
					'valid':(monitoring_crt.get_notBefore(),monitoring_crt.get_notAfter()),
				}
				
			except:
				monitoring_cert = None
		
		with node_config_lock:
			node_config['certificates']['backup'] = backup_cert
			node_config['certificates']['vpn'] = vpn_cert
			node_config['certificates']['monitoring'] = monitoring_cert
				
		# TODO - updatovat, jen pokud se neco zmenilo
		if run_node_status:
			self.realod_backup_lists = True
			#self.update_node_status()
			self.sync_config()
	
	def get_certificate_storage_path(self,file_name):
		"""docstring for get_certificate_storage_path"""
		return os.path.join(self.get('node','cert_path'),file_name)
		
	# def update_node_status(self):
# 		"""reload local config from runtime data and update config hash in node_status"""
# 		node_config_hash = hashlib.sha1()
# 		self.get_node_status()
# 		node_config_hash.update(cPickle.dumps(node_config))
#
# 		send = False
# 		with node_status_lock:
# 			if(node_status['node_config_hash'] != node_config_hash.hexdigest()):
# 				node_status['flags'] = node_status['flags'] | NODE_FLAGS_CONFIG_UPDATED
# 				node_status['node_config_hash'] = node_config_hash.hexdigest()
# 				send = True
#
# 			if(send):
# 				try:
# 					util.keepalive.send_now()
#
# 				except:
# 					pass
#
# 		del node_config_hash
	
	def get_node_status(self):
		"""update local config from runtime data, this function calls separate functions for all elements of local config"""
		self.get_network_status()
		self.get_lvm_status()
		self.get_raid_status()
		self.get_libvirt_status()
		self.get_cpuinfo()
		self.get_meminfo()
		self.get_cpuusage()
	
	def get_network_status(self):
		"""docstring for get_network_status"""
		try:
			myip = self.get('daemon','listen')
		except:
			# get ip address
			try:
				myip = subprocess.check_output(["/bin/sh","-c","ip ad sh dev brup | grep 'inet 10\.'"]).strip().split()[1].split('/')[0]
			except:
				raise Exception('unable to get ip address')
		
		hostname = socket.gethostname()
		if(hostname == 'localhost'):
			hostname = node_status['uuid'] + '.cloud-anywhere.eu'
			logger.warning('\'localhost\' hostname detected. Setting new hostname to \'{}\'.'.format(hostname))
			os.system('hostnamectl set-hostname ' + hostname)
		
		with node_status_lock:
			node_status['unicast_ip'] = myip
			
		with node_config_lock:
			node_config['unicast_ip'] = myip
			node_config['hostname'] = hostname
			
		self.get_network_interfaces()
		pass
	
	# TODO - cpuinfo a meminfo dodavat v samostatnem volani. To se bude porad menit...
	# prozatim je to profiltrovane pomoci MEMINFO_KEYS a CPUINFO_BLACKLIST_KEYS
	def get_cpuinfo(self):
		cpuinfo = open('/proc/cpuinfo').read().strip().split('\n\n')
		node_config['cpu'] = map(lambda x: dict(filter(lambda z: z[0] not in CPUINFO_BLACKLIST_KEYS,map(lambda y: (y.split(':')[0].strip(),y.split(':')[1].strip()),x.split('\n')))),cpuinfo)
		
	def get_cpuusage(self,):
		cpu_usage = None
		last_uptime = None
		domains_usage = {}
		with node_config_lock:
			cpu_usage = copy.deepcopy(node_config['cpu_usage'])
			last_uptime = copy.deepcopy(node_config['uptime'])
			for domain in node_config['libvirt']['domains']:
				domains_usage[domain] = copy.deepcopy(node_config['libvirt']['domains'][domain]['cpu_usage'])
			
		
		conn = libvirt.open('qemu:///system')
		try:
			info = conn.getInfo()
			
		except:
			return False
			
		socket_count = info[5]
		cpu_core_count = info[6]
		core_count = info[2]
		threads = info[7]
		
	
		try:
			vserver_list = map(lambda x: (x.name(), x.ID(), x.UUIDString(), x.info()[3]), conn.listAllDomains())
			conn.close()
		
		except:
			return False
	
	
		with open('/proc/uptime', 'r') as fh:
			try:
				uptime = float(fh.read().split(' ')[0])
				
			except:
				return False
		
		uptime_diff = uptime - last_uptime
		
		with open('/sys/fs/cgroup/cpuacct/cpuacct.usage', 'r') as fh:
			usage = fh.read()
		
		try:
			usage = float(usage) / 1000000000
		
		except:
			return False
		
		with open('/sys/fs/cgroup/cpuacct/cpuacct.usage_percpu', 'r') as fh:
			usage_cores = fh.read()
		
		try:
			usage_cores = map(lambda x: float(x.strip()) / 1000000000, usage_cores.strip().split(' '))
		
		except:
			return False
		
		if(uptime_diff != 0):
			cpu_usage['usage_percent'] = (usage - cpu_usage['usage']) / uptime_diff / int(core_count) * 100
			cpu_usage['usage'] = usage
			index = 0
			for usage_core in usage_cores:
				if(len(cpu_usage['cores']) <= index):
					cpu_usage['cores'].append({'usage': 0, 'usage_percent': 0})
		
				cpu_usage['cores'][index]['usage_percent'] = (usage_core - cpu_usage['cores'][index]['usage']) / uptime_diff * 100
				cpu_usage['cores'][index]['usage'] = usage_core
				index += 1
			
			for vserver in vserver_list:
				if(vserver[2] not in domains_usage):
					continue
				
				cgroup_path = '/sys/fs/cgroup/cpuacct/machine.slice/machine-qemu\\x2d%s\\x2d%s.scope/' % (vserver[1], vserver[0].replace('_', ''))
				if(not os.path.exists(cgroup_path)):
					continue
			
				with open(cgroup_path + 'cpuacct.usage', 'r') as fh:
					usage = fh.read()
		
				try:
					usage = float(usage) / 1000000000
		
				except:
					return False
			
				domains_usage[vserver[2]]['usage_percent'] = (usage - domains_usage[vserver[2]]['usage']) / uptime_diff / vserver[3] * 100
				domains_usage[vserver[2]]['usage'] = usage
			
				ls_dir = os.listdir(cgroup_path)
				ls_dir = filter(lambda x: re.match('^vcpu[0-9]+$', x) != None, ls_dir)
				ls_dir.sort()
				index = 0
				for vcpu in ls_dir:
					if(len(domains_usage[vserver[2]]['cores']) <= index):
						domains_usage[vserver[2]]['cores'].append({'usage': 0, 'usage_percent': 0})
					
					path = cgroup_path + vcpu + '/cpuacct.usage'
					with open(path, 'r') as fh:
						usage_core = fh.read()
		
					try:
						usage_core = float(usage_core) / 1000000000
		
					except:
						return False
				
					domains_usage[vserver[2]]['cores'][index]['usage_percent'] = (usage_core - domains_usage[vserver[2]]['cores'][index]['usage']) / uptime_diff * 100
					domains_usage[vserver[2]]['cores'][index]['usage'] = usage_core	
				
					index +=1
		
		with node_config_lock:
			node_config['uptime'] = uptime
			node_config['cpu_usage'] = cpu_usage
			for domain in domains_usage:
				node_config['libvirt']['domains'][domain]['cpu_usage'] = domains_usage[domain]
				
		
			
		return True

	def get_meminfo(self):
		meminfo = open('/proc/meminfo').read().strip().split('\n')
		mem = {}
		for line in meminfo:
			line = line.split(':')
			if line[0] in MEMINFO_KEYS:
				mem[line[0]] = int(line[1].strip('kB').strip())
		node_config['meminfo'] = mem
	
	def get_lvm_status(self):
		"""docstring for get_lvm_status"""
		try:
			devnull = open('/dev/null','w')
			lvs = subprocess.check_output(['lvs','--noheadings', '--binary', '--unbuffered','--units','B'],stderr=devnull, stdin=devnull).strip().split('\n')
			lvs = map(lambda x: x.split(), lvs)
			pvs = subprocess.check_output(['pvs','--noheadings', '--binary', '--unbuffered','--units','B'],stderr=devnull, stdin=devnull).strip().split('\n')
			pvs = map(lambda x: x.split(), pvs)
			vgs = subprocess.check_output(['vgs','--noheadings', '--binary', '--unbuffered','--units','B'],stderr=devnull, stdin=devnull).strip().split('\n')
			vgs = map(lambda x: x.split(), vgs)
			with node_config_lock:
				node_config['lvm']['lvs'] = lvs
				node_config['lvm']['pvs'] = pvs
				node_config['lvm']['vgs'] = vgs
				
		except:
			pass
			
	def get_raid_status(self):
		try:	
			mdstat = {}
			mdstat_fh = open('/proc/mdstat', 'r')
			data = mdstat_fh.read().split('\n', 1)[1]
			mdstat_fh.close()
			data = re.split('\n\W+\n', data)
			for raid in data:
				tmp = re.match('^(md[0-9]+) : ([a-z]+) ([a-z0-9]+) (.*)\n.*\[([U_]+)\](\n.*resync\ =\ +([0-9.]+))?', raid)
				if(tmp == None):
					continue
					
				dev_count = len(tmp.group(5))
				devs = [None] * dev_count
				for dev in tmp.group(4).split(' '):
					dev = re.match('^(.*)\[([0-9]+)\]', dev)
					if(dev == None):
						continue
						
					devs[int(dev.group(2))] = dev.group(1) 
					
				mdstat[tmp.group(1)] = {
					'status': tmp.group(2),
					'type': tmp.group(3),
					'devs': devs,
					'sync': True if(tmp.group(7) == None) else float(tmp.group(7))
				}
			
			with node_config_lock:
				node_config['mdstat'] = mdstat
	
		except:
			pass
		
	
	def get_libvirt_status(self):
		"""docstring for get_libvirt_status"""
		backup_server = None
		with node_config_lock:
			status = cPickle.loads(cPickle.dumps(node_config['libvirt'], -1))
			backup_server = copy.deepcopy(node_config['backup_server'])
			
		conn = libvirt.open('qemu:///system')
		status['libvirt_version'] = conn.getLibVersion()
		# status['node_info'] = conn.getInfo()
		domain_list = []
		domain_uuid_list = []
		for domain in conn.listDefinedDomains():
			try:
				domain_list.append(conn.lookupByName(domain))
			except:
				pass
				
		for domain in conn.listDomainsID():
			try:
				domain_list.append(conn.lookupByID(domain))
			except:
				pass
		
		domain_uuids = map(lambda x: x.UUIDString(), domain_list)
		for domain in status['domains'].keys():
			if(domain not in domain_uuids):
				del status['domains'][domain]
		
				
		
		for dom in domain_list:
			domain_uuid = dom.UUIDString()
			try:
				domain_name = virt.guest.get_name_from_uuid(domain_uuid)
				
			except:
				continue
				
			domain_uuid_list.append(domain_uuid)
			if not status['domains'].has_key(domain_uuid):
				status['domains'][dom.UUIDString()] = {
					'xml':None,
					'inactive_xml':None,
					'status':None,
					'backups': {'running': [], 'schedule': None, 'completed': None, 'pending': None},
					'cpu_usage': {
						'usage': 0,
						'usage_percent': 0,
						'cores': []
					}
				}
			
			backup_list = None
			if(backup_server['unavailable_timestamp'] == None or backup_server['unavailable_timestamp'] + 20 < time.time()):
				backup_list = BackupList(domain_name)
				if(not backup_list.is_available()):
					backup_server['unavailable_timestamp'] = time.time()
					logger.warning('Backup server is unavailable')
				
				else:
					if(backup_server['unavailable_timestamp'] != None):
						logger.info('Backup server is available')
						
					backup_server['unavailable_timestamp'] = None
					status['domains'][domain_uuid]['backups']['completed'] = backup_list
					
			if(status['domains'][domain_uuid]['backups']['schedule'] == None):	
				status['domains'][domain_uuid]['backups']['schedule'] = BackupSchedule(backup_list, domain_name)
			
			if(self.realod_backup_lists and backup_server['unavailable_timestamp'] == None):
				status['domains'][domain_uuid]['backups']['completed'].refresh()
				
			try:	
				status['domains'][domain_uuid]['xml'] = dom.XMLDesc(libvirt.VIR_DOMAIN_XML_SECURE)
				status['domains'][domain_uuid]['inactive_xml'] = dom.XMLDesc(libvirt.VIR_DOMAIN_XML_SECURE|libvirt.VIR_DOMAIN_XML_INACTIVE)
				status['domains'][domain_uuid]['status'] = dom.info()[0]
				xml = parseString(status['domains'][domain_uuid]['xml'])
				status['domains'][domain_uuid]['name'] = str(xml.getElementsByTagName('name')[0].childNodes[0].data)
			
			except:
				pass
		
		with node_config_lock:
			node_config['libvirt'] = status
			node_config['backup_server'] = backup_server
	
	def get_network_interfaces(self):
		ret = subprocess.check_output(['ip','ad','sh'])
		iface = None
		ifaces = []
		for line in ret.strip().split('\n'):
			mymatch = re.match('^[0-9]+:\s([a-z0-9]+):',line)
			if mymatch:
				# next interface
				if iface != None:
					if addr != None:
						iface['addr'].append(addr)
					ifaces.append(iface)
				iface = {'name':mymatch.group(1)}
				sp = line.split()
				iface['flags'] = sp[2].strip('<>')
				for i in range(3,len(sp),2):
					iface[sp[i]] = sp[i+1]
				addr = None
				iface['addr'] = []

			else:
				# parsujeme adresy
				line = line[4:]
				if line[0] != ' ':
					if addr != None:
						iface['addr'].append(addr)
					sp = line.split()
					if sp[0] != 'link/none':
						addr = {'type':sp[0],'addr':sp[1],'flags':[]}
					else:
						addr = {'type':sp[0],'addr':None,'flags':[]}
					i = 2
					while i < len(sp):
						if sp[i] in ['scope','brd']:
							addr[sp[i]] = sp[i+1]
							i+=2
						else:
							addr['flags'].append(sp[i])
							i+=1
				else:
					sp = line.split()
					for i in range(0,len(sp),2):
						if sp[i] in ['valid_lft', 'preferred_lft']:
							continue
						addr[sp[i]] = sp[i+1]


		if iface != None:
			if addr != None:
				iface['addr'].append(addr)
			ifaces.append(iface)
		with node_config_lock:
			node_config['interfaces'] = ifaces
			
		return
