import util,config,threading,network,time,remote,remote_api,commands,vpn,logger,backups,diagnostics,cache
import signal,sys,os

from c2padmin.virt import volume
from c2padmin.virt.guest import get_name_from_uuid
from version import __version__

logger = logger.getLogger(__name__)

global unexpectly_end
def unexpectly_end(ex, resume = False):
	file_name = '/tmp/c2padmin-' + time.strftime('%Y%m%dT%H%M%S%Z') + '.dump'
	logger.exception('uncaught excetion ({}): dump saved to {}'.format(str(ex), file_name))
	fh = open(file_name, 'w')
	fh.write(diagnostics.diagnostics.make_dump())
	fh.close()
	
	if(resume == False):
		open('/tmp/c2padmin_crash', 'w').close()
		if(threading.current_thread().name == 'C2Padmin_main'):
			dispatcher.terminate(signal.SIGTERM, None)
		
		else:
			dispatcher.graceful_terminate = True
		
	

class Dispatcher():
	"""this class dispatches all required threads and processes"""
	def __init__(self):
		self.network = network.Network()
		network.network = self.network
		threading.current_thread().name = 'C2Padmin_main'
		self.graceful_terminate = False
		self.threads = {}
		self.lock = None
		
		try:
			os.unlink('/tmp/c2padmin_crash')
			
		except:
			pass
		
		pass
	
	def terminate(self,signum,frame):
		if(self.lock != None):
			try:
				self.lock.release()
				
			except:
				pass
			
		if not config.node_config['preinstall']:
			with config.node_status_lock:
				config.node_status['flags'] |= config.NODE_FLAGS_TERMINATING
			util.keepalive.send_now()
			remote.server_instance.stop()
		config.main_thread_signal = signum
		
		cache.cache.iso_cache_purge()
		
		for key in self.threads:
			logger.info('Waiting for {} to terminate'.format(key))
			self.threads[key].join()
			
		diagnostics.diagnostics.close()
		for remote_shell_pid in util.remote_shell_proc:
			util.remote_shell_proc[remote_shell_pid].terminate()
			util.remote_shell_proc[remote_shell_pid].wait()
		if signum == signal.SIGHUP:
			import __main__ as main
			os.execlp('python', 'python', main.__file__)
		else:
			sys.exit(0)
	
	def dispatch(self):
		signal.signal(signal.SIGTERM, self.terminate)
		signal.signal(signal.SIGINT, self.terminate)
		signal.signal(signal.SIGHUP, self.terminate)
		
		logger.info('Starting c2padmin {}'.format(__version__))
		
		logger.info('Running node status')
		config.config.get_node_status()
		config.config.sync_config()
		
		network.bridge_admin = network.BridgeAdmin()
		network.bridge_admin.checkout_config()
		
		network.dhcp = network.DHCPServer()
		
		carbon_timeout = config.config.get('node', 'carbon_timeout', int)
		if not config.node_config['preinstall']:
			backups.dispatcher.dispatcher_instance = backups.dispatcher.BackupDispatcher()
		
		now = time.time()
		
		diagnostics.diagnostics = diagnostics.Diagnostics(config.config.parser.get('diagnostics', 'port'))
		
		try:
			diagnostics.diagnostics.allow_from(map(lambda x: x.strip(), config.config.parser.get('diagnostics', 'allow_from').split(',')))
			
		except:
			diagnostics.diagnostics.allow_from(['127.0.0.1/8'])
			logger.error('bad value of diagnostics.allow_from - allow from localhost enabled')
			
		try:	
			while True:
				if(self.graceful_terminate):
					dispatcher.terminate(signal.SIGTERM, None)
					
				if not (self.threads.has_key('status_loop') and self.threads['status_loop'].is_alive()):
					logger.info('Starting status loop')
					self.threads['status_loop'] = threading.Thread(target=config.config.status_loop,args=())
					self.threads['status_loop'].daemon = True
					self.threads['status_loop'].name = 'status_loop'
					self.threads['status_loop'].start()
				
				if not (self.threads.has_key('sync_config_loop') and self.threads['sync_config_loop'].is_alive()):
					logger.info('Starting sync config loop')
					self.threads['sync_config_loop'] = threading.Thread(target=config.config.sync_config_loop,args=())
					self.threads['sync_config_loop'].daemon = True
					self.threads['sync_config_loop'].name = 'sync_config_loop'
					self.threads['sync_config_loop'].start()
				
				if not config.node_config['preinstall']:
					if not (self.threads.has_key('keepalive_server') and self.threads['keepalive_server'].is_alive()):
						logger.info('Starting keepalive server')
						util.keepalive = util.Keepalive(self.network)
						self.threads['keepalive_server'] = threading.Thread(target=util.keepalive.server_loop,args=())
						self.threads['keepalive_server'].daemon = True
						self.threads['keepalive_server'].name = 'keepalive_server'
						self.threads['keepalive_server'].start()
			
					if not (self.threads.has_key('keepalive_client') and self.threads['keepalive_client'].is_alive()):
						logger.info('Starting keepalive client loop')
						self.threads['keepalive_client'] = threading.Thread(target=util.keepalive.client_loop,args=())
						self.threads['keepalive_client'].daemon = True
						self.threads['keepalive_client'].name = 'keepalive_client'
						self.threads['keepalive_client'].start()
			
					if not (self.threads.has_key('remote_server') and self.threads['remote_server'].is_alive()):
						logger.info('Starting remote handler')
						self.threads['remote_server'] = threading.Thread(target=remote.start_server,args=())
						self.threads['remote_server'].daemon = True
						self.threads['remote_server'].name = 'remote_server'
						self.threads['remote_server'].start()
			
					if not (self.threads.has_key('backup_dispatcher') and self.threads['backup_dispatcher'].is_alive()):
						logger.info('Starting backup runner')
						self.threads['backup_dispatcher'] = threading.Thread(target=backups.dispatcher.dispatcher_instance.main_loop,args=())
						self.threads['backup_dispatcher'].daemon = True
						self.threads['backup_dispatcher'].name = 'backup_dispatcher'
						self.threads['backup_dispatcher'].start()
			
					if not (self.threads.has_key('backup_scheduler') and self.threads['backup_scheduler'].is_alive()):
						logger.info('Starting backup runner')
						self.threads['backup_scheduler'] = threading.Thread(target=backups.dispatcher.dispatcher_instance.scheduler_loop,args=())
						self.threads['backup_scheduler'].daemon = True
						self.threads['backup_scheduler'].name = 'backup_scheduler'
						self.threads['backup_scheduler'].start()
					
					if config.node_config['roles']['vpn'] == 'master':
						if not (self.threads.has_key('vpn') and self.threads['vpn'].is_alive()):
							logger.info('Starting vpn process')
							vpn.vpn_instance = vpn.OpenVPN(config.config.get('node','vpn_host'))
							self.threads['vpn'] = threading.Thread(target=vpn.vpn_instance.vpn_process,args=())
							self.threads['vpn'].daemon = True
							self.threads['vpn'].name = 'vpn'
							self.threads['vpn'].start()
			
				if('cc' not in dir(config)):
				 	config.cc = commands.CommandsCaller()
				 	logger.info('Initialize commands caller')
							
				if config.node_config['roles']['api'] == 'master':
					if not (self.threads.has_key('remote_api') and self.threads['remote_api'].is_alive()):
						remote_api.api = remote_api.RemoteApi(config.config.get('node','api_address'))
						ret = remote_api.api.login(config.node_uuid_bytes,config.api_key)
						print ret
						self.threads['remote_api'] = threading.Thread(target=remote_api.client_loop,args=(self.network,))
						self.threads['remote_api'].daemon = True
						self.threads['remote_api'].name = 'remote_api'
						self.threads['remote_api'].start()
						
				
				if not (self.threads.has_key('dhcp') and self.threads['dhcp'].is_alive()):
					logger.info('Starting DHCPd process')
					self.threads['dhcp'] = threading.Thread(target=network.dhcp.maintenance_loop,args=())
					self.threads['dhcp'].daemon = True
					self.threads['dhcp'].name = 'dhcp'
					self.threads['dhcp'].start()
					
			
				snapshots_to_remove = []
				
				# Check remote shells
				terminated = []
				for remote_shell_pid in util.remote_shell_proc:
					if(util.remote_shell_proc[remote_shell_pid].poll() != None):
						logger.info('Remote shell (pid: %s) terminated with exit code %s' % (remote_shell_pid, util.remote_shell_proc[remote_shell_pid].returncode))
						terminated.append(remote_shell_pid)
						# Mozna info administraci ze je port uz volnej
				
				self.lock = config.node_config_lock
				with config.node_config_lock:
					if(len(terminated) != 0):
						util.remote_shell_proc = {pid: process for pid, process in util.remote_shell_proc.iteritems() if(pid not in terminated)}
						config.node_config['remote_shell'] = filter(lambda x: x['pid'] not in terminated, config.node_config['remote_shell'])
					
					#update info.timestamp
					if(config.node_config['info']['timestamp'] + 1 < time.time()):
						config.node_config['info']['timestamp'] = time.time()
					
					if(config.node_config['info']['state'] == config.ST_STARTING):
						config.node_config['info']['state'] = config.ST_UP	
				
				self.lock = None
				# Snapshot cleaner
		
				if(int(time.time()) != int(now) and int(time.time()) % 30 == 0): # Spust kazdych 30s
					now = time.time()
					cache_size = cache.cache.get_isos_cache_size() / 1024.0 / 1024.0 / 1024.0
					if(cache_size > 20):
						cache.cache.remove_lru_iso()
					
					snapshots = volume.get_snapshots()
					if(snapshots == None):
						logger.warning('Snapshot cleaner - get snapshot failed')
					
					else:
						to_delete = []
						self.lock = config.node_config_lock
						with config.node_config_lock:
							for snapshot, timestamp in config.node_config['active_snapshots'].iteritems():
								if(snapshot not in snapshots):
									to_delete.append(snapshot)
									#del config.node_config['active_snapshots'][snapshot]
									continue
					
								if(int(now) % 600 == 0 and timestamp + (2 * 24 * 60 * 60) < now):
									if(timestamp + (10) < now):
										logger.warning('Snapshot {} is older than 2 days'.format(snapshot))
					
							for snapshot in to_delete:
								del config.node_config['active_snapshots'][snapshot]
					
							for snapshot in snapshots:
								if snapshot not in config.node_config['active_snapshots']:
									logger.info('Snapshot {} is no longer needed, will be removed'.format(snapshot))
									snapshots_to_remove.append(snapshot)
				
						self.lock = None
						
				for snapshot in snapshots_to_remove:
					volume.remove_volume(snapshot)
				
				self.lock = config.node_status_lock
				with config.node_status_lock:
					# Check disconnected nodes
					to_remove = []
					carbon_dests = config.config.get_carbon_destinations()
				
					for uuid in config.node_status['disconnect_node_data']:
						if(time.time() > config.node_status['disconnect_node_data'][uuid]['timestamp'] + carbon_timeout):
							if(config.node_status['disconnect_node_data'][uuid]['unicast_ip'] + ':2004' in carbon_dests):
								logger.info('Carbon timeout for %s exceeded' % uuid)
								del carbon_dests[carbon_dests.index(config.node_status['disconnect_node_data'][uuid]['unicast_ip'] + ':2004')]
								to_remove.append(uuid)
				
					if(len(to_remove)):			
						config.config.set_carbon_destinations(carbon_dests)
					
					config.node_status['disconnect_node_data'] = {uuid: data for uuid, data in config.node_status['disconnect_node_data'].iteritems() if(uuid not in to_remove)}
				
				self.lock = None
					
				time.sleep(0.5)
					
		except Exception as ex:
			unexpectly_end(ex)
