import threading,time,cPickle,traceback,copy,subprocess,os,re,json,random,hashlib,signal,struct,socket
import util,config,remote,logger

import c2padmin

logger = logger.getLogger(__name__)

def remote_fcall(fname, data, transferred_bytes = None):
	"""docstring for remote_fcall"""
	# pomocni funkce pro clone_server a migrate_server
	# na vzdalenem nodu spusti pozadovanou funkci s prichozimi datami pres funkci fcall
	# vstup json, povinne dst_nod_ip, func, data
	try: c = remote.Client(data["dst_nod_ip"])
	except:	raise
	try: nc = c.fcall(fname,[data])
	except:	raise
	return nc

def get_universe():
	"""docstring for get_universe"""
	# tahle vec obsahuje cely vesmir, rekl jirka :-)
	return copy.deepcopy(network.get_nodes())

class Network():
	"""
		this class maintains network configuration in sync using multicast keepalive facility and remote calls for data synchronization
		
	"""
	def __init__(self):
		self.config = {}
		self.config_lock = threading.Lock()
		self.nodes = {config.node_config['uuid']:config.node_config}
		self.keep_alive_queue_lock = threading.Lock()
		self.keep_alive_queue = []
		self.hosts = {}
		self.hosts_hash = None
		
		th = threading.Thread(target=self.maintenance_loop, args=())
		th.daemon = True
		th.start()
		
		update_network_th = threading.Thread(target = self.process_update_network)
		update_network_th.daemon = True
		update_network_th.start()
	
	def maintenance_loop(self):
		"""perform maintenance periodically"""
		try:
			while True:
				if config.main_thread_signal != None:
					# TODO
					return
				
				self.maintenance()
				
				time.sleep(config.config.get('network','maintenance_loop_interval',float))
		
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
			
	def process_update_network(self):
		try:
			while True:
				if config.main_thread_signal != None:
					return
					
				try:
					with self.keep_alive_queue_lock:
						data = self.keep_alive_queue.pop(0)
						
				except:
					time.sleep(0.1)
					continue
					
				self.update_network(data)
				
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
			
	def process_keepalive(self, data):
		if(data['uuid'] in self.config):
			self.config[data['uuid']]['last_update'] = time.time()
			
		with self.keep_alive_queue_lock:
			if(len(filter(lambda x: x['uuid'] == data['uuid'], self.keep_alive_queue)) == 0):
				self.keep_alive_queue.append(data)
				
			else:
				self.keep_alive_queue = map(lambda x: data if(x['uuid'] == data['uuid']) else x, self.keep_alive_queue)
					
	
	def get_api_master_uuid(self):
		for uuid,node in self.nodes.iteritems():
			if(node['roles']['api'] == 'master'):
				return uuid
				
		return None
	
	def maintenance(self):
		"""
			perform maintenance of network status record
			currently this only drops timeouted records
		"""
		with self.config_lock:
			# drop timeouted
			for uuid in self.nodes.keys():
				if uuid == config.node_config['uuid']:
					continue
				if(config.node_config['roles']['api'] == 'master'):
					if(self.nodes[uuid]['info']['state'] == config.ST_DOWN and config.node_status['sync_network_timestamp'] - 1 > self.nodes[uuid]['info']['timestamp']):
						del self.nodes[uuid]
						del self.config[uuid]
						
				else:
					if(self.nodes[uuid]['info']['state'] == config.ST_DOWN and config.node_status['master_call_timestamp'] - 1 > self.nodes[uuid]['info']['timestamp']):
						del self.nodes[uuid]
						del self.config[uuid]
				
			for uuid in self.config.keys():
				item = self.config[uuid]
				if(self.nodes[uuid]['info']['state'] == config.ST_DOWN):
					continue
					
				if 'last_update' in item and time.time() > (item['last_update']+config.config.get('network','keepalive_timeout',float)):
					with config.node_status_lock:
						config.node_status['disconnect_node_data'][uuid] = {'timestamp': time.time(), 'unicast_ip': self.nodes[uuid]['unicast_ip']}
						
					self.nodes[uuid]['info']['state'] = config.ST_DOWN
					self.nodes[uuid]['info']['timestamp'] = time.time()
						
					logger.info('Timeouted %s' % (uuid,))
					
	
	def check_master_status(self):
		try:
			with config.master_status_lock:
				with config.node_status_lock:
					# check if there is master in the network
					for role in config.node_config['roles'].keys():
						if config.node_config['roles'][role] == 'disabled':
							continue
						next_role = False
						for uuid in self.nodes.keys():
							if(self.nodes[uuid]['roles'][role] == 'master' and self.nodes[uuid]['info']['state'] == config.ST_UP and uuid != config.node_config['uuid']):
								if config.node_config['roles'][role] == 'master':
									# we have a collision, need to sort this out
									# TODO
									raise Exception('We have two {} masters in the network'.format(role))
									
								else:
									# somebody else is master, yay
									config.node_status['flags'] &= ~config.NODE_FLAGS_MASTER_CANDIDATE[role]
									next_role = True
									break
									
						if next_role:
							continue
						# no master in the network are we master?
						if config.node_config['roles'][role] == 'master':
							# we are master, nothing to do
							continue
						else:
							# ok. so no master so far. Let's try to become one
							if config.node_status['flags'] & config.NODE_FLAGS_MASTER_CANDIDATE[role]:
								# check other candidates and back out if other candite with better priority exists
								stay_candidate = True
								try:
									local_priority = config.node_config['roles_prio'][role]
								except:
									local_priority = 0xffffffff
								for uuid in self.nodes.keys():
									if uuid != config.node_config['uuid'] and (self.config[uuid]['flags'] & config.NODE_FLAGS_MASTER_CANDIDATE[role]):
										try:
											remote_priority = self.nodes[uuid]['roles_prio'][role]
										except:
											# we don't know remote priority yet. Ignore this one
											continue
										if remote_priority <= local_priority:
											logger.info('Better remote candidate for {}. Backing out. Local prio {}, remote prio {}'.format(role,local_priority,remote_priority))
											# remote server has higher priority. back out
											stay_candidate = False
											break
								if not stay_candidate:
									# unset candidate flag and continue to other role
									config.node_status['flags'] &= ~config.NODE_FLAGS_MASTER_CANDIDATE[role]
									continue
								if (time.time() - config.master_candidature_time[role]) > config.config.get('network','role_switch_grace_period',float):
									# grace period passed. Take over
									# notify all that have master candidate flags in their keepalive message
									notify_host_list = []
									with self.config_lock:
										for uuid in self.nodes.keys():
											if uuid != config.node_config['uuid'] and (self.config[uuid]['flags'] & config.NODE_FLAGS_MASTER_CANDIDATE[role]):
												notify_host_list.append(self.config[uuid]['unicast_ip'])
									skip_this_time = False
									for host in notify_host_list:
										try:
											if remote.proxy_call(host).master_is_calling(role) == False:
												# remote refused our setting of master status. try next time
												logger.info('Remote denied our becoming {} master'.format(role))
												skip_this_time = True
												break
										except:
											logger.warning(traceback.format_exc())
											pass
									if skip_this_time:
										continue
									config.node_status['flags'] &= ~config.NODE_FLAGS_MASTER_CANDIDATE[role]
									config.node_config['roles'][role] = 'master'
									logger.info('Became {} master'.format(role))
									continue
								else:
									# waiting for grace period
									continue
							else:
								become_candidate = True
								try:
									local_priority = config.config.get('daemon','{}_priority'.format(role),int)
								except:
									local_priority = 0
								# let's see other candidates and their priorities
								with self.config_lock:
									for uuid in self.nodes.keys():
										if uuid != config.node_config['uuid'] and (self.config[uuid]['flags'] & config.NODE_FLAGS_MASTER_CANDIDATE[role]):
											if local_priority == 0:
												become_candidate = False
												break
											try:
												remote_priority = self.nodes[uuid]['roles_prio'][role]
											except:
												# we don't know remote priority yet. Ignore this one
												continue
											if remote_priority <= local_priority:
												# remote server has higher priority. back out
												logger.info('Not becoming candidate for {}. Local prio {}, remote prio {}'.format(role,local_priority,remote_priority))
												become_candidate = False
												break
								if become_candidate:
									if local_priority == 0:
										local_priority = random.randint(0x10000,0xffffffff)
									logger.info('Setting {} candidate flag, local prio {}'.format(role,local_priority))
									config.node_status['flags'] |= config.NODE_FLAGS_MASTER_CANDIDATE[role]
									config.node_config['roles_prio'][role] = local_priority
									config.master_candidature_time[role] = time.time()
		except IOError:
			# pass if config.master_status_lock cannot be acquired
			pass
		
	def get_nodes(self):
		"""docstring for get_nodes"""
		# chceme kopii bez objektu a podobnych cicaradicek. takze to prozeneme jsonem
		ret = json.loads(json.dumps(self.nodes,cls=util.ExtendedEncoder))
		for node_uuid in ret.keys():
			if node_uuid == config.node_config['uuid']:
				ret[node_uuid]['node_config_hash'] = config.node_status['node_config_hash']
		return ret
		
	def refresh_hosts(self):
		hosts = {}
		for uuid in self.nodes.keys():
			if('hostname' in self.nodes[uuid] and 'unicast_ip' in self.nodes[uuid]):
				hosts[self.nodes[uuid]['hostname']] = self.nodes[uuid]['unicast_ip']
				
		hosts_hash = hashlib.sha1()
		hosts_hash.update(cPickle.dumps(hosts))
		hosts_hash = hosts_hash.hexdigest()
		if(hosts_hash != self.hosts_hash):
			logger.info('Hosts list updated')
			self.hosts = hosts
			self.hosts_hash = hosts_hash
			try:
				fh = open('/etc/hosts', 'r')
				lines = fh.readlines()
				fh.close()
				fh = open('/etc/hosts', 'w')
				for line in lines:
					line = line.strip()
					tmp = line.split('#')
					if(len(tmp) == 2 and tmp[1] == 'C2P_local_node'):
						continue
						
					fh.write(line + '\n')
					
				for host in hosts:
					fh.write(hosts[host] + '\t' + host + '\t#C2P_local_node\n')
					
				fh.close()
			
			except:
				pass
				
	def update_network(self,data):
		"""
			update network status record with node data
		"""
		logger.debug('Running update network')
		try:
			if config.get_network_identifier(data['uuid']) == config.node_config['net_identifier'] and data['uuid'] != config.node_config['uuid']:
				with self.config_lock:
					if data['flags'] & config.NODE_FLAGS_TERMINATING:
						if self.config.has_key(str(data['uuid'])):
							del self.config[str(data['uuid'])]
					else:
						new_carbon_node = False
						if not self.config.has_key(str(data['uuid'])):
							logger.info('joining '+str(data['uuid']))
							
							
						self.config[str(data['uuid'])] = data
						if not self.nodes.has_key(str(data['uuid'])) or self.nodes[str(data['uuid'])]['node_config_hash'] != data['node_config_hash']:
							logger.info('config of %s changed, updating' % (str(data['uuid'])))
							nc = self.fetch_node_config(data['uuid'], data['node_config_hash'])
							if nc != None:
								logger.info('Got config of {} hash {}'.format(str(data['uuid']),data['node_config_hash']))
								if(not self.nodes.has_key(str(data['uuid']))):
									new_carbon_node = True
									
								self.nodes[str(data['uuid'])] = nc
								with config.node_config_lock:
									if nc['bridge_config']['serial'] > config.node_config['bridge_config']['serial']:
										config.node_config['bridge_config']['data'] = copy.deepcopy(nc['bridge_config']['data'])
										config.node_config['bridge_config']['serial'] = nc['bridge_config']['serial']
								bridge_admin.checkout_config()
							else:
								logger.info('unable to fetch config, dropping '+str(data['uuid']))
								try:
									del self.config[str(data['uuid'])]
									del self.nodes[str(data['uuid'])]
								except:
									pass
								return
								
								
						else:
							logger.debug('config of %s unchanged' % (str(data['uuid'])))
							
						if(self.nodes[data['uuid']]['roles']['api'] == 'master'):
							with config.node_status_lock:
								config.node_status['master_call_timestamp'] = time.time()
						
						if(new_carbon_node):
							with config.node_status_lock:
								if(data['uuid'] in config.node_status['disconnect_node_data']):
									del config.node_status['disconnect_node_data'][data['uuid']]
								
							carbon_dests = config.config.get_carbon_destinations()
							if(carbon_dests != False and self.nodes[data['uuid']]['unicast_ip'] + ':2004' not in carbon_dests):
							
								carbon_dests.append(self.nodes[data['uuid']]['unicast_ip'] + ':2004')
								config.config.set_carbon_destinations(carbon_dests)
								
							
						self.config[str(data['uuid'])]['last_update'] = time.time()
					
				self.maintenance()
			
			self.check_master_status()
			self.refresh_hosts()
			
		except:
			raise
			pass

	def fetch_node_config(self,node_uuid, node_config_hash):
		"""
			fetch remote node config and update network
		"""
		try:
			nc = remote.proxy_call(self.config[str(node_uuid)]['unicast_ip']).get_node_config()
			logger.debug('Node config:'+str(nc))
			nc['node_config_hash'] = node_config_hash
			return nc
		except:
			return None

class BridgeAdmin(object):
	"""docstring for BridgeAdmin"""
	def __init__(self):
		self.eoip = {
			'pid': None,
			'active_uplink': 'brup',
			'unicast_ip': config.node_config['unicast_ip'],
			'bind_ip': None
		}
		if config.config.get('network','public',int):
			try:
				route = subprocess.check_output(['ip', 'route', 'show']).split('\n')
				
			except:
				raise Exception('Couldn\'t get ip route')
			
			for rec in route:
				if(rec[:7] != 'default'):
					continue
					
				self.eoip['active_uplink'] = rec.split('dev ', 1)[1].split()[0].strip()
				break
				
			try:
				fh = open('/var/run/eoip-eoip.cfg', 'r')
				self.eoip['pid'] = int(fh.read().strip())
				fh.close()
				
			except:
				pass
				
						
			iface = filter(lambda x: 'master' in x and x['master'] == self.eoip['active_uplink'] and x['state'] == 'UP', config.node_config['interfaces'])
			if(len(iface) != 1):
				raise Exception('Couldn\'t determine uplink interface')
				
			self.uplink_iface = iface[0]['name']
			self.vxlan_iface = 'breoip'
			self.public = True
			
			try:
				addresses = subprocess.check_output(['ip', 'addr', 'show', 'dev', self.eoip['active_uplink']]).split('\n')
				
			except:
				raise Exception('Couldn\'t get ip route')
			
			for address in addresses:
				address = address.strip().split()
				if(address[0] == 'inet' and address[-1] == self.eoip['active_uplink']):
					self.eoip['bind_ip'] = address[1].split('/')[0]
					break
					
				continue
			
		else:
			self.uplink_iface = 'brup'
			self.vxlan_iface = 'brup'
			self.public = False
			fh = open('/sys/devices/virtual/net/%s/bridge/multicast_querier' % (self.vxlan_iface,),'w')
			fh.write('1')
			fh.close()
			
		# kdyz neni zapnuty qurrier v linux bridgi, tak prestava po 5 minutach chodit multicast
		# viz: http://unix.stackexchange.com/questions/132604/br0-causes-to-drop-multicast-connection-after-5-minutes
		
		self.brctl_hash = None
		self.ip_link_hash = None
		self.vlan_serial = -1
		self.connections = []
		self.bridges = []
		self.vpn_active = False
		self.config = None
		self.peer_ips = []
		self.active_peers = []
		self.local_range = False
		self.apply_config_lock = threading.Lock()
		
		pass
		
	def __local_range(self, addr, prefix):
		try:
			split_addr = addr.split('.')
			split_addr = map(lambda x: int(x), split_addr)
			
		except:
			raise Exception('Bad IP format')
			
		try:
			prefix = int(prefix)
			
		except:
			raise Exception('Bad prefix value')
			
		if(len(split_addr) != 4):
			raise Exception('Bad IP format')
			
		if(prefix < 22 or prefix >= 31):
			raise Exception('Bad prefix value')
			
		if(split_addr[0] == 10):
			return True
			
		elif(split_addr[0] == 192 and split_addr[1] == 168):
			return True
			
		elif(split_addr[0] == 172 and split_addr[1] >= 16 and  split_addr[1] <= 31):
			return True
			
		return False

	def checkout_config(self):
		"""docstring for checkout_config"""
		run_apply = False
		with self.apply_config_lock:
			with config.node_config_lock:
				if config.node_config['bridge_config']['serial'] > self.vlan_serial:
					self.config = config.node_config['bridge_config']['data']
					run_apply = True
			
				if(self.public):
					tmp = None
					if(config.node_config['roles']['api'] != 'master'):
						master_uuid = network.get_api_master_uuid()
						if(master_uuid != None):
							tmp = network.nodes[master_uuid]['peer_ips']
						
					else:
						tmp = config.node_config['peer_ips']

					if(tmp == None):
						api_engine = c2padmin.remote_api.RemoteApi(config.config.get('node','api_address'))
						ret = api_engine.login(config.node_uuid_bytes,config.api_key)
						tmp = api_engine.get_locality_server_ips()				
				
					if(tmp != self.peer_ips):
						self.peer_ips = tmp
						self.__make_eoip_config()
						self.__setup_eoip()
						run_apply = True
				
			if(run_apply):
				retVal = self.apply_config()
				time.sleep(1)
				return retVal

	def get_serial(self):
		"""docstring for get_serial"""
		return self.vlan_serial

	def __apply_iface_config(self,iface_config):
		if int(iface_config['vlan_id']) == 0:
			iface_config['vlan_id'] = 1
		if int(iface_config['vlan_id']) > 1:
			bridge_name = 'bx{}.{}'.format(iface_config['vxlan_id'].lower(),iface_config['vlan_id'])
		else:
			bridge_name = 'bx{}'.format(iface_config['vxlan_id'].lower())

		if(bridge_name not in self.bridges):
			self.bridges.append(bridge_name)

		if self.vpn_active:
			vpn_name = 'vpntun.{}'.format(iface_config['vlan_id'])
		
		if iface_config['local_vlan_id'] == None:
			uplink_name = 'vx{}.{}'.format(iface_config['vxlan_id'].lower(),iface_config['vlan_id'])
		else:
			if int(iface_config['local_vlan_id']) > 1:
				uplink_name = '{}.{}'.format(self.uplink_iface,iface_config['local_vlan_id'])
			else:
				assert not self.public, 'Untagged vlan in public cloud! This should not happen!'
				uplink_name = 'vup'
		
		self.__check_iface(uplink_name)
		self.__check_iface(bridge_name)
		if(len(iface_config['subnets'])):
			self.__assign_subnets(bridge_name, iface_config['subnets'])
			
		if self.vpn_active:
			self.__check_iface(vpn_name)
			self.__check_bridge_uplink(bridge_name, [uplink_name,vpn_name])
			
		else:
			self.__check_bridge_uplink(bridge_name, [uplink_name])
			
		pass
		
	def __assign_subnets(self, iface_name, subnets):
		for subnet in subnets:
			os.system('ip address add {} dev {}'.format(subnet['subnet'], iface_name))
			tmp = subnet['subnet'].split('/')
			if(self.__local_range(tmp[0], tmp[1])):
				os.system('ip address add {} dev {}'.format(subnet['gateway'], iface_name))
				self.local_range = True
		
	def __make_eoip_config(self):
		if(not os.path.exists('/etc/eoip')):
			os.mkdir('/etc/eoip', 0755)
		
		file_conten = ''
		index = 0
		for peer in self.peer_ips:
			if(config.node_config['uuid'] != peer['uuid'] and peer['ip'] != None):
				file_conten += '[zeoip{}]\nid={}\ndst={}\n\n'.format(index, index + 100, peer['ip'])
				index += 1
				
		fh = open('/etc/eoip/eoip.cfg', 'w')
		fh.write(file_conten)
		fh.close()
		
	def __setup_eoip(self):
		if(self.eoip['pid'] != None):
			try:
				os.kill(self.eoip['pid'], signal.SIGINT)
				os.unlink('/var/run/eoip-eoip.cfg')
				
			except:
				pass
		
		if(len(filter(lambda x: x['ip'] != None, self.peer_ips)) > 1):
			logger.info('Starting EoIP')
			subprocess.call(['eoip', '/etc/eoip/eoip.cfg', self.eoip['bind_ip']])	
			time.sleep(1)
			try:
				fh = open('/var/run/eoip-eoip.cfg', 'r')
				self.eoip['pid'] = int(fh.read().strip())
				fh.close()
			
			except:
				logger.error('EoIP starting failed')
				return False
		
			logger.info('EoIP started')
			br_status = self.__parse_bridge_status(self.vxlan_iface)
			if(br_status == None):
				if(os.system('brctl addbr {}'.format(self.vxlan_iface)) != 0):
					raise Exception('Cannot create {} bridge'.format(self.vxlan_iface))
				
				fh = open('/sys/devices/virtual/net/%s/bridge/multicast_querier' % (self.vxlan_iface,),'w')
				fh.write('1')
				fh.close()
				br_status = self.__parse_bridge_status(self.vxlan_iface)
				
			for index in range(len(filter(lambda x: x['ip'] != None, self.peer_ips)) - 1):
				if('zeoip{}'.format(index) not in br_status['slaves']):
					if(os.system('brctl addif {} zeoip{}'.format(self.vxlan_iface, index)) != 0):
						raise Exception('Cannot add if zeoip{} to {} bridge'.format(index, self.vxlan_iface))
					
					if(os.system('ip link set mtu 2000 dev zeoip{}'.format(index)) != 0):
						raise Exception('Cannot set mtu on zeoip{} if'.format(index))
		
	
			if(os.system('ip link set up {}'.format(self.vxlan_iface)) != 0):
				raise Exception('Cannot set link up on {}'.format(self.vxlan_iface))
	
			if(os.system('ip link set mtu 2000 dev {}'.format(self.vxlan_iface)) != 0):
				raise Exception('Cannot set mtu on {} if'.format(self.vxlan_iface))
		
			unicast_ip_full = None
			unicast_ip_dev = None
			try:
				data = subprocess.check_output(['ip', 'ad', 'sh'])
				data = filter(lambda x: len(re.findall(re.escape(self.eoip['unicast_ip']), x)) != 0, re.split('\n[0-9]+:\ ', data))[0]
				unicast_ip_dev = re.match('^(.*?):', data, re.M).group(1)
				unicast_ip_full = re.findall(re.escape(self.eoip['unicast_ip']) + '/[0-9]+', data)[0]

			except:
				unicast_ip_full = self.eoip['unicast_ip'] + '/21'
				unicast_ip_dev = None
		
			if(unicast_ip_dev != None and unicast_ip_dev != self.vxlan_iface):
				if(os.system('ip ad del {} dev {}'.format(unicast_ip_full, unicast_ip_dev)) != 0):
					raise Exception('Cannot del ip {} from if {}'.format(unicast_ip_full, unicast_ip_dev))
		
			if(unicast_ip_dev != self.vxlan_iface):
				if(os.system('ip ad add {} dev {}'.format(unicast_ip_full, self.vxlan_iface)) != 0):
					raise Exception('Cannot add ip {} on if {}'.format(unicast_ip_full, self.vxlan_iface))
		
			os.system('ip ro del 239.0.0.0/8')
			if(os.system('ip ro add 239.0.0.0/8 dev {} src {}'.format(self.vxlan_iface, self.eoip['unicast_ip'])) != 0):
				raise Exception('Cannot add multicast route')
		

	def __check_iface(self,name):
		if os.system('ip link sh {} >/dev/null 2>&1'.format(name)) == 0:
			# TODO - check parameters
			return True
		else:
			mymatch = re.match('^(bx[0-9a-z]{6})(\.([0-9]+))?$',name)
			if mymatch != None:
				os.system('brctl addbr {}'.format(name))
				os.system('brctl stp {} on'.format(name))
				os.system('ip link set dev {} up'.format(name))
				return True
				
			mymatch = re.match('^(vx([0-9a-z]{6}))(\.([0-9]+))?$',name)
			if mymatch != None:
				if mymatch.group(3) != None:
					r = self.__check_iface(mymatch.group(1))
					if r and os.system('ip link add link vx{0} vx{0}.{1} mtu 1500 type vlan proto 802.1Q id {1} && ip link set up vx{0}.{1}'.format(mymatch.group(2),mymatch.group(4))) == 0:
						return True
					else:
						return False
				else:
					if os.system('ip link add vx{0} mtu 1500 type vxlan id 0x{0} group {1} dev {3} dstport {2} && ip link set up vx{0}'.format(mymatch.group(2), config.config.get('network','vxlan_group'), config.config.get('network','vxlan_port'),self.vxlan_iface)) == 0:
						return True
					else:
						return False
			mymatch = re.match('^{0}\.([0-9]+)$'.format(self.uplink_iface),name)
			if mymatch != None:
				if os.system('ip link add link {1} {1}.{0} mtu 1500 type vlan proto 802.1Q id {0} && ip link set up {1}.{0}'.format(mymatch.group(1),self.uplink_iface)) == 0:
					return True
				else:
					return False
			mymatch = re.match('^vpntun\.([0-9]+)$',name)
			if mymatch != None:
				if os.system('ip link add link vpntun vpntun.{0} mtu 1500 type vlan proto 802.1Q id {0} && ip link set up vpntun.{0}'.format(mymatch.group(1))) == 0:
					return True
				else:
					return False
			if name == 'vup':
				assert not self.public, 'Cannot use veth in public cloud!'
				if os.system('ip link add vup type veth peer name vupb && ip link set up vup && brctl addif {} vupb && ip link set up vupb'.format(self.uplink_iface)) == 0:
					# do bridge infrastruktury nepoustet multicast provoz vxlan a keepalive z uplinku
					os.system('ebtables -t nat -A POSTROUTING -o vupb --ip-dst {address} -p IPv4 --ip-proto udp --ip-dport {port} -j DROP'.format(address=config.config.get('network','multicast_group'),port=config.config.get('network','multicast_port')))
					os.system('ebtables -t nat -A POSTROUTING -o vupb --ip-dst {address} -p IPv4 --ip-proto udp --ip-dport {port} -j DROP'.format(address=config.config.get('network','vxlan_group'),port=config.config.get('network','vxlan_port')))
					os.system('ebtables -t nat -A POSTROUTING -o vupb -p 802_1Q -j DROP')
					return True
				else:
					return False
			raise Exception('Unimplemented interface type {}'.format(name))

	def __check_bridge_uplink(self,bridge_name,uplink_name_list):
		br_data = self.__parse_bridge_status(bridge_name)
		for uplink_name in uplink_name_list:
			if not uplink_name in br_data['slaves']:
				os.system('brctl addif {} {}'.format(bridge_name,uplink_name))
			for slave in br_data['slaves']:
				if re.match('^(vup|{}|vx|vpntun)'.format(self.uplink_iface),slave) != None and slave not in uplink_name_list:
					os.system('brctl delif {} {}'.format(bridge_name,slave))
				
		return True

	def __parse_bridge_status(self,bridge_name=None):
		proc = None
		if bridge_name == None:
			proc = subprocess.Popen(['brctl','show'], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		else:
			proc = subprocess.Popen(['brctl','show',bridge_name], stdout = subprocess.PIPE, stderr = subprocess.PIPE)

		data, stderr = proc.communicate()
		
		if(proc.returncode != 0 or len(stderr.strip()) != 0):
			return None
			
		data = data.strip().split('\n')[1:]
		# parse line by line
		br = None
		bridges = {}
		for line in data:
			if line[0] not in ['\t',' ']:
				if br != None:
					bridges[br['name']] = br
				sp = line.split()
				br = {
					'name':sp[0],
					'id':sp[1],
					'stp':sp[2],
					'slaves':[]
				}
				if len(sp) > 3:
					br['slaves'].append(sp[3])
			else:
				br['slaves'].append(line.strip())
			if br != None:
				bridges[br['name']] = br
		if bridge_name != None:
			return bridges.get(bridge_name)
		else:
			return bridges
	
	def __apply_backbone(self,):
		pass
	
	def apply_config(self,config=None):
		if config != None:
			self.config = config
			
		if self.config == None:
			return
		
		if self.public:
			self.__apply_backbone()

		self.bridges = []
		for iface_config in self.config:
			self.__apply_iface_config(iface_config)

		bridge_list = self.__parse_bridge_status()
		for bridge in bridge_list:
			if(bridge not in self.bridges and bridge != self.uplink_iface):
				self.remove_bridge(bridge, bridge_list[bridge]['slaves'])
				
		if(self.local_range):
			os.system('iptables -t nat -C POSTROUTING -o brup -j MASQUERADE || iptables -t nat -A POSTROUTING -o brup -j MASQUERADE')
			os.system('sysctl -w net.ipv4.ip_forward=1')

		if len(self.config):
			self.vlan_serial = int(self.config[0]['vlan_serial'])
			
		pass

	def update_status(self):
		# parse current bridge state and update object variables
		pass

	def add_bridge(self):
		# based on local_vlan_id either set up uplink_iface.vlan_id as uplink OR vxlan.vlan_id interfaces
		pass

	def remove_bridge(self,name, slaves = []):
		# check if uplink interface is still in use after bridge removal, eventually clean up
		if os.system('ip link sh {} > /dev/null 2>&1'.format(name)) != 0:
			return True
			
		if re.match('^(bx[0-9a-z]{6})(\.([0-9]+))?$',name) != None:
			if os.system('ip link set {} down'.format(name)) != 0:
				raise Exception('Could not disable bridge {}'.format(name))
				
			if os.system('brctl delbr {}'.format(name)) != 0:
				raise Exception('Could not delete bridge {}'.format(name))
				
			for iface in slaves:
				if(iface == self.uplink_iface):
					continue
					
				if os.system('ip link delete {}'.format(iface)) != 0:
					raise Exception('Could remove interface {}'.format(iface))
				
		pass

	def add_connection(self):
		# add veth connection between bridges (never connect brup with bx bridge containing vx interface)
		pass

	def remove_connection(self):
		# remove veth connection
		pass
		
		
class DHCPServer(object):
	
	def __init__(self):
		self._dhcpd_running = False
		self._bridges = {}
		
	def _ip2int(self, addr):
	    return struct.unpack("!I", socket.inet_aton(addr))[0]


	def _int2ip(self, addr):
	    return socket.inet_ntoa(struct.pack("!I", addr))
		
	def _local_range(self, addr, prefix):
		try:
			split_addr = addr.split('.')
			split_addr = map(lambda x: int(x), split_addr)
			
		except:
			raise Exception('Bad IP format')
			
		try:
			prefix = int(prefix)
			
		except:
			raise Exception('Bad prefix value')
			
		if(len(split_addr) != 4):
			raise Exception('Bad IP format')
			
		if(prefix < 22 or prefix >= 31):
			raise Exception('Bad prefix value')
			
		if(split_addr[0] == 10):
			return True
			
		elif(split_addr[0] == 192 and split_addr[1] == 168):
			return True
			
		elif(split_addr[0] == 172 and split_addr[1] >= 16 and  split_addr[1] <= 31):
			return True
			
		return False
			
		
	def _make_config(self, data):
		try:
			content = ''
			for subnet_id, subnet in data['subnets'].iteritems():
				network = subnet['subnet'].split('/')
				prefix = int(network[1])
				network = network[0]
				if(subnet['reserved'] == ''):
					tmp = []
					
				else:
					tmp = subnet['reserved'].split(',')
					
				reserved = []
				while len(tmp):
					minimum = [2**(32 - prefix), 2**(32 - prefix)]
					
					for item in tmp:
						val = int(item.split('-')[0])
						if(val < minimum[0]):
							minimum = [val, item]
					
					reserved.append(minimum[1])
					tmp.remove(minimum[1])
				
				pool_list = []
				pool_low = 1
				pool_high = (32 - prefix)**2 - 1
				for item in reserved:
					item = item.split('-')
					if(len(item) == 1):
						item = [int(item[0]), int(item[0])]
					
					else:
						item[0] = int(item[0])
						item[1] = int(item[1])
					
					if(pool_low == item[0]):
						pool_low = item[1] + 1
					
					elif(pool_low < item[0]):
						pool_list.append([pool_low, item[0] - 1])
						pool_low = item[1] + 1
					
					else:
						raise Exception('Bad pool definition')
					
				if(pool_low < pool_high):
					pool_list.append([pool_low, pool_high])
				
				content += 'subnet {} netmask {} {{\n'.format(network, subnet['netmask'])
				content += '\tdefault-lease-time 7200;\n'
				content += '\tdeny unknown-clients;\n'
				content += '\toption routers {};\n'.format(subnet['gateway'])
				content += '\toption subnet-mask {};\n'.format(subnet['netmask'])
				content += '\toption domain-name-servers {};\n'.format(subnet['dns_servers'])
				for pool in pool_list:
					content += '\trange {} {};\n'.format(self._int2ip(self._ip2int(network) + pool[0]),self._int2ip(self._ip2int(network) + pool[1]))
				
				content += '}\n'
			
			if(len(data['hosts']) != 0):
				for vserver, host in data['hosts'].iteritems():
					network = data['subnets'][host['id_subnet']]['subnet'].split('/')[0]
					ip_address = self._int2ip(self._ip2int(network) + int(host['subnet_increment']))
					content += 'host {}_{} {{\n'.format(vserver, str(host['id_subnet']) + '_' + str(host['subnet_increment']))
					content += '\thardware ethernet {};\n'.format(host['mac'].lower())
					content += '\tfixed-address {};\n'.format(ip_address)
					content += '}\n'
			
			fh = open('/etc/dhcp/dhcpd.conf', 'w')
			fh.write(content)
			fh.flush()
			fh.close()
		
			return True
			
		except Exception as ex:
			logger.error('Make config failed: ' + str(ex))
			return False
			
	def _start_dhcpd(self):
		try:
			subprocess.check_output(['systemctl', 'status', 'dhcpd'])
			return False
		
		except:
			os.system('systemctl start dhcpd')
			self._dhcpd_running = True
			return True
			
	def _stop_dhcpd(self):
		try:
			subprocess.check_output(['systemctl', 'status', 'dhcpd'])
			os.system('systemctl stop dhcpd')
			self._dhcpd_running = False
			return True
		
		except:
			return False
			
	def _status_dhcpd(self):
		try:
			subprocess.check_output(['systemctl', 'status', 'dhcpd'])
			self._dhcpd_running = True
			return True
		
		except:
			self._dhcpd_running = False
			return False
			
	def _clean_network_setting(self):
		try:
			bridges = subprocess.check_output(['brctl', 'show']).split('\n')[1:]
			bridges = map(lambda x: x.split()[0], filter(lambda x: x[:2] == 'bx', bridges))
			for bridge in bridges:
				os.system('ip -f inet address flush dev {}'.format(bridge))
				
			os.system('iptables -t nat -D POSTROUTING -o brup -j MASQUERADE')
			os.system('sysctl -w net.ipv4.ip_forward=0')
			
		except:
			logger.error('Cleanup DHCP setting failed')
			raise
		
	
	def maintenance_loop(self):
		"""perform maintenance periodically"""
		try:
			self._clean_network_setting()
			while True:
				self._status_dhcpd()
				if(config.main_thread_signal != None):
					if(self._dhcpd_running):
						logger.info('Stoping dhcpd')
						self._stop_dhcpd()
						self._clean_network_setting()
						
					return
				
				if(config.node_config['roles']['api'] != 'master'):
					if(self._dhcpd_running):
						logger.info('Stoping dhcpd')
						self._stop_dhcpd()
						self._clean_network_setting()
						
					time.sleep(1)
					continue
				
				with config.node_config_lock:
					data = config.node_config['dhcp']
				
				if(not data['reload_config']):
					time.sleep(1)
					continue
				
				if(len(data['config']['subnets']) == 0 or len(data['config']['hosts']) == 0):
					logger.info('No subnet or hosts defined - stoping dhcpd')
					self._stop_dhcpd()
					self._clean_network_setting()
					with config.node_config_lock:
						data = config.node_config['dhcp']['reload_config'] = False
						
					time.sleep(1)
					continue
				
				if(config.node_config['bridge_config']['serial'] != bridge_admin.vlan_serial):
					time.sleep(1)
					continue
					
				with bridge_admin.apply_config_lock:	
					logger.info('Reloading dhcpd configuration')
					logger.info('Stoping dhcpd')
					self._stop_dhcpd()
					self._clean_network_setting()
				
					res = self._make_config(data['config'])
					if(res):
						logger.info('/etc/dhcp/dhcpd.conf prepared')
						time.sleep(5)
					
					else:
						logger.error('Preparing of /etc/dhcp/dhcpd.conf failed')
						time.sleep(30)
						continue
					
					res = self._start_dhcpd()
					if(res):
						logger.info('dhcpd started')
					
					else:
						logger.error('Starting dhcpd failed')
						time.sleep(30)
						continue
				
					with config.node_config_lock:
						data = config.node_config['dhcp']['reload_config'] = False
					time.sleep(1)
		
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
		
		
		
		