import rfoo,cPickle,Queue,threading
import config,network,util,commands,cert,logger

import subprocess, os, virt

import c2padmin

logger = logger.getLogger(__name__)

remote_connections_lock = threading.Lock()
remote_connections = {}
server_instance = None

class RemoteHandler(rfoo.BaseHandler):
	"""
		handler class fro unicast remote server
		
		define methods that will be called by a remote client
	"""
	def __init__(self, *args, **kwargs):
		rfoo.BaseHandler.__init__(self, *args, **kwargs)
	
	def ping(self):
		"""docstring for ping"""
		return True
	
	def call__(self,caller_uuid,fname,params):
		"""wrapper function to handle pickling possible encryption"""
		args,kwargs = cPickle.loads(params)
		fn = self.__getattribute__(fname)
		self.caller_uuid = caller_uuid
		return cPickle.dumps(fn(*args,**kwargs))
	
	def get_node_config(self):
		with config.node_status_lock:
			config.node_status['flags'] &= ~config.NODE_FLAGS_CONFIG_UPDATED
			
		with config.node_config_lock:
			return config.node_config
	
	def save_ssh_keys(self, pub, priv):
		with config.node_config_lock:
			config.node_config['ssh_keys'] = {
				'pub': pub,
				'priv': priv
			}
			
		return config.config.save_ssh_keys()
	
	def enqueue_commands(self,commands):
		for command in commands:
			config.cc.put(command)
		return True
		
	def fetch_completed_commands(self):
		"""docstring for fetch_completed_commands"""
		return commands.fetch_completed_commands()
	
	def master_is_calling(self,role):
		"""docstring for master_is_calling"""
		try:
			with config.master_status_lock:
				if config.node_config['roles'][role] == 'master':
					logger.info('Denied remote request for becoming master!')
					return False
				with config.node_status_lock:
					try:
						config.node_status['flags'] &= ~config.NODE_FLAGS_MASTER_CANDIDATE[role]
					except KeyError:
						return False
				logger.info('Another host is master for {}, reverting to slave'.format(role))
				with config.node_config_lock:
					config.node_config['roles'][role] = 'slave'
				with network.network.config_lock:
					if network.network.nodes.has_key(self.caller_uuid):
						network.network.nodes[self.caller_uuid]['roles'][role] = 'master'
				return True
		except IOError:
			return False

	def new_certificate_request(self,feature):
		"""docstring for new_certificate_request"""
		return cert.new_request(feature)
		
	def save_certificate(self,crt,feature):
		"""docstring for save_certificate"""
		return cert.save_certificate(crt,feature)
	
	def reload_certificates(self):
		"""docstring for reload_certificates"""
		return config.config.reload_certificates()
	
	def fcall(self,fname,args):
		"""docstring for fcall"""
		# spusti na vzdalenem nodu funkci fname s prichozimi argumentami
		flist = fname.split('.')
		modname = '.'.join(flist[:-1])
		fname = flist[-1]
		if modname == 'virt':
			module = virt
		elif modname == 'virt.guest':
			module = virt.guest
		elif modname == 'util':
			module = util
		else:
			return 'Invalid function call: unimplemented module or permission denied'
		fun = getattr(module,fname)
		if fun == None:	return 'Invalid function call: function unknown'
		try: nc = fun(*args)
		except Exception, e: nc = e
		return nc
			
class Client(object):
	"""
		client class for rfoo proxy
		
		this class provides data pickling, calling remote function and unpickling the result
		
		
		call using
		Client(remote_host).function_name(*args,**kwargs)
	"""
	def __init__(self, remote_host):
		self._remote_host = remote_host
		self.__conn = rfoo.InetConnection().connect(host=self._remote_host,port=config.config.get('network','unicast_port',int))
		self._lock = threading.Lock()
	
	def __getattr__(self, name):
		def fn(*args, **kwargs):
			with self._lock:
				ret = rfoo.Proxy(self.__conn).call__(config.node_config['uuid'],name,cPickle.dumps([args,kwargs]))
			return cPickle.loads(ret)
		
		return fn
	
def start_server():
	"""start rfoo unicast server"""
	try:
		global server_instance
		server_instance = rfoo.InetServer(RemoteHandler)
		server_instance.start(host=config.node_status['unicast_ip'], port=config.config.get('network','unicast_port',int))
		
	except Exception as ex:
		c2padmin.dispatcher.unexpectly_end(ex)

def proxy_call(remote_ip,recurse = True):
	"""docstring for fname"""
	global remote_connections,remote_connections_lock
	with remote_connections_lock:
		if remote_ip not in remote_connections:
			client = Client(remote_ip)
			remote_connections[remote_ip] = client
		else:
			client = remote_connections[remote_ip]
	try:
		client.ping()
	except:
		with remote_connections_lock:
			try:
				del remote_connections[remote_ip]
			except:
				pass
		if recurse:
			return proxy_call(remote_ip,recurse=False)
	return client