import requests,json,time,datetime
from base64 import b64encode

# import for type definitions
import uuid,backups

import config,remote,util,commands,network,cert,logger,cPickle,hashlib
import c2padmin

logger = logger.getLogger(__name__)


class RemoteApi(object):
	"""docstring for RemoteApi"""
	def __init__(self,api_base_address):
		self._api_base_address = api_base_address
		self._params = {'sid':None}
	
	def use_sid(self,sid):
		"""docstring for use_sid"""
		previous_sid = self._params['sid']
		self._params['sid'] = sid
		try:
			ret = self.__getattr__('is_logged')()
			if ret == True:
				return True
		except:
			pass
		self._params['sid'] = previous_sid
		return False
		
	def login(self,uuid_bytes,api_key):
		"""docstring for login"""
		try:
			login = self.__getattr__('login')(b64encode(uuid_bytes),b64encode(api_key))
			
		except:
			
			return False
			
		self._params['sid'] = login['sid']
		return True
	
	def logout(self):
		"""docstring for logout"""
		ret = self.__getattr__('logout')()
		self._params['sid'] = None
		return True
	
	def __getattr__(self, name):
		def fn(*args):
			"""Call method on server. handle pickling possible encryption"""
			logger.info('Calling remote api: {}'.format(name))
			call_data = {'fn': name,'args':json.dumps(args,cls=util.ExtendedEncoder)}
			if self._sid != None:
				call_data['sid'] = self._params['sid']
			r = requests.post(self._api_base_address,data=call_data)
			try:
				if r.status_code == requests.codes.ok:
					return r.json()
					
				else:
					raise Exception()
					
			except:
				if(r.status_code == 503):
					raise Exception('maintenance')
					
				raise Exception('Function %s returned error code %s with text: %s' % (name,r.status_code,r.content))
		
		return fn
	
		

def client_loop(netw):
	"""docstring for client_loop"""
	global api
	next_try = None
	try:
		while True:
			if config.node_config['roles']['api'] != 'master' or config.main_thread_signal != None:
				try:
					del api
				except:
					pass
				return
				
			if(next_try != None):
				if(next_try > time.time()):
					time.sleep(1)
					continue
					
				else:
					next_try = None
					
			try:
				nodes = netw.get_nodes()
				sync_reply = api.sync({u:v['node_config_hash'] for u,v in nodes.iteritems()})
				with config.node_config_lock:
					try:
						if config.config.get('network','public',int):
							config.node_config['peer_ips'] = api.get_locality_server_ips()
							
						if config.node_config['bridge_config']['serial'] < int(sync_reply['vlan_serial']):
							config.node_config['bridge_config']['data'] = api.get_vlan_config()
							config.node_config['bridge_config']['serial'] = int(sync_reply['vlan_serial'])
							
						tmp = api.get_dhcp_config()
						dhcp_hash = hashlib.sha1()
						dhcp_hash.update(cPickle.dumps(tmp))
						if(tmp != None and dhcp_hash.hexdigest() != config.node_config['dhcp']['hash']):
							config.node_config['dhcp']['hash'] = dhcp_hash.hexdigest()
							config.node_config['dhcp']['config'] = tmp
							config.node_config['dhcp']['reload_config'] = True
								
					except:
						pass
			
				network.bridge_admin.checkout_config()
				# check certificate status based on features and request new if desirable
				for feature in ['backup','vpn','monitoring']:
					if feature in sync_reply['features']:
						for node_uuid in nodes:
							if nodes[node_uuid]['certificates'][feature] == None:
								if node_uuid == config.node_config['uuid']:
									req = cert.new_request(feature)
									signing_result = api.sign_certificate_request(req,feature)
									if signing_result[0]:
										crt = signing_result[1]
										cert.save_certificate(crt,feature)
										config.config.reload_certificates()
									else:
										logger.debug('Signing of certificate failed for server '+node_uuid)
								else:
									print "calling {} for certificate request".format(node_uuid)
									c = remote.Client(netw.config[node_uuid]['unicast_ip'])
									req = c.new_certificate_request(feature)
									signing_result = api.sign_certificate_request(req,feature)
									if signing_result[0]:
										crt = signing_result[1]
										c.save_certificate(crt,feature)
										c.reload_certificates()
									else:
										logger.debug('Signing of certificate failed for server '+node_uuid)
				
				for node_uuid in nodes:
					if(nodes[node_uuid]['ssh_keys']['priv'] == None or nodes[node_uuid]['ssh_keys']['pub'] == None):
						ssh_pair = api.get_ssh_key_pair()
						if(type(ssh_pair) == list and ssh_pair[0] == True):
							if node_uuid == config.node_config['uuid']:
								logger.info('Setting ssh keys for server ' + node_uuid)
								with config.node_config_lock:
									config.node_config['ssh_keys'] = {
										'pub': ssh_pair[1]['pub_key'],
										'priv': ssh_pair[1]['priv_key']
									}
								
								res = config.config.save_ssh_keys()
								if(not res):
									logger.error('Setting ssh keys for server ' + node_uuid + ' failed')
								
							else:
								logger.info('Setting ssh keys for server ' + node_uuid)
								c = remote.Client(netw.config[node_uuid]['unicast_ip'])
								res = c.save_ssh_keys(ssh_pair[1]['pub_key'], ssh_pair[1]['priv_key'])
								if(not res):
									logger.error('Setting ssh keys for server ' + node_uuid + ' failed')
								
				
				# ret['commands'] obsahuje pripadne commandy
				if sync_reply['need_config_update']:
					logger.debug('Updating network info')
					ret = api.sync_network_status(nodes)
					if(ret):
						with config.node_status_lock:
							config.node_status['sync_network_timestamp'] = time.time()
				
				for u in sync_reply['commands'].keys():
					if u == config.node_config['uuid']:
						cmds = sync_reply['commands'][u]
						logger.debug('Received %d local commands' % (len(cmds),))
						if len(cmds):
							for command in cmds:
								config.cc.put(command)
					else:
						logger.debug('Received %d remote commands for %s' % (len(sync_reply['commands'][u]),u))
						try:
							c = remote.proxy_call(netw.config[str(u)]['unicast_ip']).enqueue_commands(sync_reply['commands'][u])
						except:
							# nepodarilo se zaradit commandy? Nevadi. Prijdou znovu.
							pass
				with netw.config_lock:
					cmds = []
					for u in netw.nodes:
						if netw.nodes[u]['commands_in_queue']:
							if u == config.node_config['uuid']:
								cmds += commands.fetch_completed_commands()
							else:
								c = remote.Client(netw.config[str(u)]['unicast_ip'])
								cmds += c.fetch_completed_commands()
					if len(cmds):
						try:
							api.commands_reply(cmds)
						
						except:
							for cmd in cmds:
								print "Vracim command k pozdejsimu zpracovani"
								config.cc.put_complete(cmd)
							
							raise
							
			except Exception as ex:
				# unable to communicate with remote api
				# going to slave status
				if(str(ex) == 'maintenance'):
					logger.warning('remote API maintenance, sleeping for 60s')
					next_try = time.time() + 60
					
				else:	
					with config.node_config_lock:
						if not config.node_config['preinstall']:
							config.node_config['roles']['api'] = 'slave'
				
					logger.error('unable to communicate with remote api: ' + str(ex))
					c2padmin.dispatcher.unexpectly_end(ex, True)
				#raise
			time.sleep(config.config.get('network','remote_api_interval',float))
			
	except Exception as ex:
		unexpectly_end(ex)