#!/usr/bin/python
# -*- coding: utf-8 -*-
VERSION = '0.1'

import socket, threading, time, select, Queue, json, struct
import c2padmin

class Socket_RPC(object):
	def __init__(self, addr, port, timeout = 10, max_connections = 20, data_size_limit = 1024*1024):
		self._addr = addr
		self._port = port
		self._timeout = timeout
		self._max_connections = max_connections
		self._data_size_limit = data_size_limit
		
		self._serve = False
		self._main_thread = None
		self._server_socket = None
		
		self._sockets = []
		self._conn_data = {}
		self._workers = {}
		
		self._login_function = None
		
		self._func_map = {}
		self._ip_list = [[0, 0]]
		
	def bind_login(self, func):
		if(func == None):
			self._login_function = None
			
		if(not callable(func)):
			raise TypeError('expected function')
			
		self._login_function = func
		
	def bind(self, func, name = None):
		if(not callable(func)):
			raise TypeError('expected function')
			
		if(name == None):
			name = func.__name__
			
		if(name in self._func_map):
			raise RuntimeError('function is registered')
			
		self._func_map[name] = func
		
	def allow_from(self, ip_list = []):
		self._ip_list = []
		for ip in ip_list:
			ip = ip.split('/', 1)
			if(len(ip) != 2):
				prefix = 32
				
			else:
				prefix = int(ip[1])	
				
			mask = (2**32 - 1) ^ (2**(32-prefix) - 1)
			ip = ip[0]
			self._ip_list.append([struct.unpack("!I", socket.inet_aton(ip))[0] & mask, mask])
				
		return True
	
	def _is_ip_allow(self, ip):
		ip = struct.unpack("!I", socket.inet_aton(ip))[0]
		return len(filter(lambda x: x[0] == ip & x[1], self._ip_list)) != 0
		
	def serve_forever(self):
		if(self._main_thread != None):
			raise RuntimeError('server is running')
		
		self._server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self._server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self._server_socket.bind((self._addr, int(self._port)))
		self._server_socket.listen(5)
		self._server_socket.setblocking(0)
		
		self._main_thread = threading.Thread(target = self._listen)
		self._main_thread.start()
		return True
		
	def is_running(self):
		return self._main_thread != None
	
	def close(self):
		self._serve = False
		try:
			self._server_socket.close()
		
		except:
			pass
		
		
		for sock in self._sockets:
			self._conn_data[sock.fileno()]['connect'] = False
			if(sock.fileno() in self._workers):
				self._workers[sock.fileno()].join()
				del self._workers[sock.fileno()]
		
			self._sockets.remove(sock)
			del self._conn_data[sock.fileno()]
				
			try:
				sock.close()

			except:
				pass
		
		self._main_thread.join()
		self._server_socket = None
		self._sockets = None
		self._main_thread = None
		return True
		
	def _listen(self):
		self._serve = True
		try:
			while self._serve:
				try:
					rs, ws, es = select.select([self._server_socket], [], [], 0.01)
				
				except socket.error as error:
					self.close()
					break
				
				if(len(rs) != 0):
					try:
						sock, addr = self._server_socket.accept()
						if(len(self._sockets) >= self._max_connections or not self._is_ip_allow(addr[0])):
							sock.close()
										
						else:
							sock.setblocking(0)
							sock.sendall(self._greeting())
							self._sockets.append(sock)
							self._conn_data[sock.fileno()] = {
								'sock': sock,
								'connect': True,
								'wait_on_result': False,
								'auth': False if(self._login_function != None) else True,
								'rx_data': Queue.Queue(),
								'tx_data': Queue.Queue(),
								'last_activity': time.time(),
							}
							
					except:
						pass
				
				try:
					rs, ws, es = select.select(self._sockets, self._sockets, [], 0.01)
				
				except socket.error as error:
					rs = []
					ws = []
				
				for sock in rs:
					try:
						data = sock.recv(1024)	
				
					except:
						data = ''
					
					if(len(data) == 0):
						self._conn_data[sock.fileno()]['connect'] = False
					
					else:
						self._conn_data[sock.fileno()]['rx_data'].put(data)
						self._conn_data[sock.fileno()]['last_activity'] = time.time()
					
				for sock in ws:
					if(not self._conn_data[sock.fileno()]['connect']):
						continue
				
					try:
						data = self._conn_data[sock.fileno()]['tx_data'].get(False)
					
					except:
						continue
				
					try:
						data += '\r\n.\r\n'
						sock.sendall(data)
						self._conn_data[sock.fileno()]['last_activity'] = time.time()
							
					except:
						self._conn_data[sock.fileno()]['connect'] = False
							
				for sock in self._sockets:
					if(self._conn_data[sock.fileno()]['last_activity'] + self._timeout < time.time() and self._conn_data[sock.fileno()]['tx_data'].empty() and self._conn_data[sock.fileno()]['wait_on_result'] == False):
						self._conn_data[sock.fileno()]['connect'] = False
					
					if(not self._conn_data[sock.fileno()]['connect']):
						if(sock.fileno() in self._workers):
							self._workers[sock.fileno()].join()
							del self._workers[sock.fileno()]
				
						self._sockets.remove(sock)
						del self._conn_data[sock.fileno()]
						
						try:
							sock.close()

						except:
							pass
						
					elif(not self._conn_data[sock.fileno()]['rx_data'].empty() and self._conn_data[sock.fileno()]['connect']):
						if(sock.fileno() not in self._workers or not self._workers[sock.fileno()].is_alive()):
							self._workers[sock.fileno()] = threading.Thread(target = self._worker, args = (sock.fileno(),))
							self._workers[sock.fileno()].start()
					
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)					
				
	def _worker(self, fileno):
		rbuffer = ''
		commnad = {}
		function = None
		args = None
		try:
			while self._conn_data[fileno]['connect']:
				if(len(rbuffer) > self._data_size_limit):
					self._conn_data[fileno]['connect'] = False
					continue
				
				try:
					rbuffer += self._conn_data[fileno]['rx_data'].get(False)
				
				except:
					time.sleep(0.01)
					continue
				
				
				if(rbuffer.find('\r\n.\r\n') == -1):
					time.sleep(0.01)
					continue
				
				commnad, rbuffer = rbuffer.split('\r\n.\r\n')
				try:
					commnad = json.loads(commnad)
				
				except:
					self._error(fileno, 'bad data format')
					continue
				
				if('func' not in commnad):
					self._error(fileno, 'bad data format')
					continue
				
				if('args' in commnad):
					args = commnad['args']
				
				else:
					args = None
				
				if(not self._conn_data[fileno]['auth'] and commnad['func'] != 'login'):
					self._error(fileno, 'authorization required')
					continue
				
				if(commnad['func'] == 'login'):
					function = self._login_function
				
				elif(commnad['func'] in self._func_map):
					function = self._func_map[commnad['func']]
				
				else:
					self._error(fileno, 'function is not registered')
					continue
			
				self._conn_data[fileno]['wait_on_result'] = True
				try:
					if(type(args) == dict):
						response = function(**args)
				
					elif(type(args) == list):
						response = function(*args)
				
					elif(args == None):
						response = function()
				
					else:
						response = function(args)
					
				except Exception as ex:
					self._call_error(fileno, str(ex))
					self._conn_data[fileno]['wait_on_result'] = False
					continue
				
				if(commnad['func'] == 'login'):
					if(response):
						self._conn_data[fileno]['auth'] = True
						self._ok(fileno, 'login successfully')
					
					else:
						self._conn_data[fileno]['auth'] = False
						self._error(fileno, 'login failed')
					
				else:
					self._ok(fileno, response)
				
				self._conn_data[fileno]['wait_on_result'] = False
				
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
			
		
	def _error(self, fileno, desc):
		self._conn_data[fileno]['tx_data'].put(json.dumps({'state': 'error', 'desc': desc}))
		
	def _call_error(self, fileno, desc):
		self._conn_data[fileno]['tx_data'].put(json.dumps({'state': 'call_error', 'desc': desc}))
		
	def _ok(self, fileno, response):
		self._conn_data[fileno]['tx_data'].put(json.dumps({'state': 'ok', 'response': response}))
		
	def _greeting(self):
		return 'Socket_RPC v' + VERSION + '\r\n'
		