import threading,cPickle,time,struct,socket,uuid,zlib,subprocess,datetime,threading,os,json,binascii,traceback,tempfile,fcntl,httplib,ConfigParser,signal
import logger
from inspect import getframeinfo, stack
from version import __version__
import mcast,config,util,virt,backups,commands
import c2padmin
logger = logger.getLogger(__name__)
remote_shell_proc = {}

class LoggingLock(object):
	def __init__(self,name):
		self._name = name
		self._lock = threading.Lock()
	
	def __enter__(self):
		self.acquire()
	
	def __exit__(self,type,value,traceback):
		"""docstring for __exit__"""
		self.release()
	
	def locked(self):
		return self._lock.locked()
		
	def acquire(self,blocking=True):
		"""docstring for acquire"""
		caller = getframeinfo(stack()[2][0])
		logger.debug('ACQ - %s - %s:%d' % (self._name, caller.filename, caller.lineno))
		return self._lock.acquire()
		
	def release(self):
		"""docstring for release"""
		caller = getframeinfo(stack()[2][0])
		logger.debug('REL - %s - %s:%d' % (self._name, caller.filename, caller.lineno))
		return self._lock.release()

class LockNB(object):
	"""docstring for LockNB"""
	def __init__(self):
		super(LockNB, self).__init__()
		self._lock_file = tempfile.TemporaryFile()
	
	def __enter__(self):
		self.acquire()
	
	def __exit__(self,type,value,traceback):
		"""docstring for __exit__"""
		self.release()
	
	def acquire(self):
		"""docstring for acquire"""
		return fcntl.flock(self._lock_file,fcntl.LOCK_EX | fcntl.LOCK_NB)
	
	def release(self):
		"""docstring for release"""
		return fcntl.flock(self._lock_file,fcntl.LOCK_UN)

class SelfUpdate(object):
	def check(self, clean_cache):
		if(clean_cache):
			proc = subprocess.Popen(['/usr/bin/yum', 'clean', 'all'], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
			stdout, stderr = proc.communicate()
			if(proc.returncode != 0):
				if(stderr.strip() != ''):
					raise Exception('yum error: {}; exited with {}'.format(stderr, proc.returncode))
			
				else:
					raise Exception('yum error: exited with {}'.format(proc.returncode))
			
		proc = subprocess.Popen(['/usr/bin/yum', '-q', 'check-update', 'c2padmin'], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		stdout, stderr = proc.communicate()
		if(proc.returncode == 0):
			return {'update_available': False, 'latest_version': __version__}
		
		elif(proc.returncode == 100):
			stdout = stdout.strip().split()
			return {'update_available': True, 'latest_version': stdout[1]}
		
		else:
			if(stderr.strip() != ''):
				raise Exception('yum error: {}; exited with {}'.format(stderr, proc.returncode))
			
			else:
				raise Exception('yum error: exited with {}'.format(proc.returncode))
	
	def update(self):
		proc = subprocess.Popen(['/usr/bin/yum', '-q', '-y', 'update', 'c2padmin'], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
		stdout, stderr = proc.communicate()
		if(proc.returncode == 0):
			open('/tmp/c2padmin_restart', 'w').close()
			return True
		
		else:
			if(stderr.strip() != ''):
				raise Exception('yum error: {}; exited with {}'.format(stderr, proc.returncode))
			
			else:
				raise Exception('yum error: exited with {}'.format(proc.returncode))

class ExtendedEncoder(json.JSONEncoder):
	"""docstring for ExtendedEncoder"""
	def default(self,obj):
		"""docstring for default"""
		if isinstance(obj,uuid.UUID):
			return str(obj)
		elif isinstance(obj,backups.types.BackupSchedule):
			return obj.get_dict()
		elif isinstance(obj,backups.types.BackupList):
			return obj.get_list(['custom_name'])
		elif isinstance(obj,backups.types.BackupJob):
			return obj.get_dict()
		elif isinstance(obj,datetime.time) or isinstance(obj,datetime.datetime) or isinstance(obj,commands.CommandProgress):
			return str(obj)
		return json.JSONEncoder.default(self, obj)
	

class Logger():
	"""
		main logging facility
	"""
	def __init__(self):
		self.lock = threading.Lock()
		pass
	def debug(self,data):
		if config.config.get('daemon','log_level') not in ['debug']:
			return
		with self.lock:
			print 'Debug: ' + str(data)
	def info(self,data):
		if config.config.get('daemon','log_level') not in ['debug','info']:
			return
		with self.lock:
			print 'Info: ' + str(data)
	def warning(self,data):
		if config.config.get('daemon','log_level') not in ['debug','info','warning']:
			return
		with self.lock:
			print 'Warning: ' + str(data)
	def error(self,data):
		if config.config.get('daemon','log_level') not in ['debug','info','warning','error']:
			return
		with self.lock:
			print 'Error: ' + str(data)

class Keepalive():
	"""
		this class provides environment for multicast keepalive network
		it distributes local status (uuid, unicast address, config hash) to multicast group
		it receives these packets and updates network config appropriately
	"""
	def __init__(self, network):
		self.server = mcast.McastServer()
		self.server.bind()
		self.network = network
		self._send_now = False
		self._lock = threading.Lock()
		self._last_sent = 0
	
	def send_now(self):
		"""docstring for send_now"""
		with self._lock:
			self._send_now = True
		return True
	
	def server_loop(self):
		"""receive packet and trigger network update"""
		try:
			while True:
				if config.main_thread_signal != None:
					# TODO
					return
				recv = self.server.recv(4096,timeout=0.1)
				if recv == None:
					continue
					
				data = self.parse_datagram(recv)
				if data != None:
					self.network.process_keepalive(data)
					
		except Exception as ex:		
			c2padmin.dispatcher.unexpectly_end(ex)
			
	
	def parse_datagram(self,data):
		"""parse multicast UDP payload struct"""
		logger.debug('Multicast datagram length: %d' % (len(data),))
		try:
			data = struct.unpack('<16s4s20sQ',data)
			data = {
				'uuid':str(uuid.UUID(bytes=data[0])),
				'unicast_ip':socket.inet_ntoa(data[1]),
				'node_config_hash':binascii.hexlify(data[2]),
				'flags':data[3],
			}
			return data
			
		except:
			return None
	
	def compose_datagram(self,data):
		"""create multicast UDP payload"""
		return struct.pack('<16s4s20sQ', uuid.UUID(data['uuid']).bytes, socket.inet_aton(data['unicast_ip']), binascii.unhexlify(data['node_config_hash']), data['flags'])
	
	def client_loop(self):
		"""send status packet to peers using multicast"""
		try:
			while True:
				if config.main_thread_signal != None:
					# TODO
					return
				send_now = False
				with self._lock:
					if self._send_now:
						send_now = True
						self._send_now = False
				
				if send_now or (time.time() > (self._last_sent+config.config.get('network','keepalive_interval',float))):
					sock = mcast.McastClient()
				
					with config.node_status_lock:
						datagram = self.compose_datagram(config.node_status)
					sock.send(datagram)
					sock.close()
					self._last_sent = time.time()
				time.sleep(0.01)
				
		except Exception as ex:		
			c2padmin.dispatcher.unexpectly_end(ex)

class UHTTPConnection(httplib.HTTPConnection):
	"""Subclass of Python library HTTPConnection that uses a unix-domain socket.
	"""
	
	def __init__(self, path,timeout=None):
		httplib.HTTPConnection.__init__(self, 'localhost')
		self.path = path
		self.timeout = timeout
 
	def connect(self):
		sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
		sock.settimeout(self.timeout)
		sock.connect(self.path)
		self.sock = sock

self_update_inst = SelfUpdate()

def check_update(clean_cache = False, progress = None):
	return self_update_inst.check(clean_cache)
	
def self_update(progress = None):
	return self_update_inst.update()
	
def update_restart_check(progress = None):
	if(not os.path.exists('/tmp/c2padmin_restart')):
		return True
		
	raise c2padmin.commands.VPSBusy(10)

def set_root_pass(password, progress=None):
	cmd='/usr/bin/passwd'
	user = 'root'
	option = '--stdin'

	p = subprocess.Popen([cmd, user, option], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	log = p.communicate(u'%(p)s\n' % { 'p': password })

	if p.returncode != 0:
		raise Exception('`passwd` failed: %d \n %s' % (p.returncode, log))
		
	return True
		
def remote_shell(ip, port, progress = None):
	with config.node_config_lock:
		if(len(filter(lambda x: x['port'] == port, config.node_config['remote_shell'])) != 0):
			raise Exception('Port already in use')
			
		cmd = '/usr/bin/ssh'
		opt = '-R:' + str(port) + ':127.0.0.1:22'
		identity_file = '-i/root/.ssh/id_rsa'
		user = '-lrescue'
		p = subprocess.Popen([cmd, identity_file, opt, user, ip], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
		config.node_config['remote_shell'].append({'pid': p.pid, 'port': port})
		remote_shell_proc[p.pid] = p
		
	return p.pid
	
def terminate_remote_shell(pid, progress = None):
	if(pid in remote_shell_proc):
		try:
			remote_shell_proc[pid].terminate()

		except OSError:
			raise Exception('Couldn\'t terminate process %s' % (pid,))
			
		return True
		
	else:
		raise Exception('couldn\'t find remote shell with PID %s' % (pid, ))
	
def copy_iso(data, progress = None):
	if(os.path.exists(data['path'])):
		return True
		
	copy_file({'src_addr': data['src_addr'], 'src_file': data['path'], 'dst_file': data['path']})
	c2padmin.cache.cache.add_iso_record(data['url'], data['size'], data['last_modified'], data['path'])
	return True
	
def copy_file(data, progress = None):
	if(os.path.exists(data['dst_file'])):
		return True
	
	p = subprocess.Popen(['/usr/bin/scp', data['src_addr'] + ':' + data['src_file'], data['dst_file']], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	log = p.communicate()
	
	if(p.returncode != 0):
		raise Exception('`scp` failed: %d \n %s' % (p.returncode, log))
		
	return True

def assign_to_network(uuid,progress=None):
	parser = ConfigParser.ConfigParser(allow_no_value=True)
	parser.read(['/etc/c2padmin/node.ini'])
	parser.set('node','uuid',uuid)
	
	ini_file_w = open('/etc/c2padmin/node.ini','w')
	parser.write(ini_file_w)
	ini_file_w.close()
	
	os.kill(os.getpid(),signal.SIGHUP)
	return True
	
	
def public_reinstall(progress = None):
	if(config.config.get('network', 'public', int) == 0):
		logger.warning('Couldn\'t reinstall non public server')
		raise Exception('Couldn\'t reinstall non public server')
	
	with config.node_config_lock:
		main_interface = filter(lambda x: 'master' in x and x['master'] == 'bond0',config.node_config['interfaces'])
		
	mac_addr = filter(lambda x: 'type' in x and x['type'] == 'link/ether', main_interface[0]['addr'])
	if(os.system('ipmitool chassis bootdev pxe') != 0):
		logger.error('Couldn\'t set boot device to PXE')
		raise Exception('Couldn\'t set boot device to PXE')
	
	logger.info('Boot device setting to PXE')
	if(os.system('/usr/sbin/shutdown -r 2') != 0):
		logger.error('Couldn\'t schedule rebbot')
		raise Exception('Couldn\'t schedule rebbot')
	
	logger.info('Reboot will be processed in 2 minutes')
	with config.node_config_lock:
		config.node_config['info']['state'] = config.ST_REINSTALL
		
	return {'mac_addr': mac_addr[0]['addr'], 'complete': 0}
