import os,subprocess, time

import random, string
import c2padmin

VG_NAME='HostVG'
STRIPE_SIZE='1m'

def get_lv(lv_name,units='g'):
	"""docstring for get_lv"""
	try:
		devnull = os.open('/dev/null',os.O_WRONLY)
		lv = subprocess.check_output(['lvs','-o','lv_name,lv_size','--noheadings', '--binary', '--unbuffered','--units',units,'%s/%s' % (VG_NAME,lv_name)],stderr=devnull, stdin=devnull).strip().split()
		os.close(devnull)
		lv[1] = float(lv[1].strip('hHbBsSkKmMgGtTpPeE').replace(',', '.'))
	except:
		lv = None
	return lv

def get_vg(units='g'):
	"""docstring for get_vg"""
	devnull = os.open('/dev/null',os.O_WRONLY)
	vg = subprocess.check_output(['vgs','-o','vg_name,vg_size,vg_free','--noheadings', '--binary', '--unbuffered','--units',units,VG_NAME],stderr=devnull, stdin=devnull).strip().split()
	os.close(devnull)
	vg[1] = float(vg[1].strip('hHbBsSkKmMgGtTpPeE').replace(',', '.'))
	vg[2] = float(vg[2].strip('hHbBsSkKmMgGtTpPeE').replace(',', '.'))
	return vg

def get_pvs(units='g'):
	"""docstring for get_vg_config"""
	devnull = os.open('/dev/null',os.O_WRONLY)
	pvs = map(lambda x: x.split(), subprocess.check_output(['pvs','-o','vg_name,pv_name,pv_size,pv_free','--noheadings', '--binary', '--unbuffered','--units',units],stderr=devnull, stdin=devnull).strip().split('\n'))
	os.close(devnull)
	pvs = [x for x in pvs if x[0] == VG_NAME]
	pvs = map(lambda x: x[1:],pvs)
	
	for i in range(len(pvs)):
		pvs[i][1] = float(pvs[i][1].strip('hHbBsSkKmMgGtTpPeE').replace(',', '.'))
		pvs[i][2] = float(pvs[i][2].strip('hHbBsSkKmMgGtTpPeE').replace(',', '.'))
	
	return pvs
	
def get_snapshots():
	try:
		devnull = os.open('/dev/null',os.O_WRONLY)
		lvs = map(lambda x: x.strip(), subprocess.check_output(['/usr/sbin/lvs', '-o', 'lv_name,lv_attr', '--noheadings', '--unbuffered'],stderr=devnull, stdin=devnull).split('\n'))
		os.close(devnull)
		
	except:
		return None
	
	retVal = []
	for lv in lvs:
		lv = lv.strip().split()
		if(len(lv) != 2 or len(lv[1]) < 1 or lv[1][0] != 's'):
			continue
			
		retVal.append(lv[0].strip())
	
	return retVal

def remove_volume(lv_name):
	"""docstring for remove_volume"""
	# smaze lv disk
	with c2padmin.config.node_config_lock:
		if(lv_name in c2padmin.config.node_config['active_snapshots']):
			del c2padmin.config.node_config['active_snapshots'][lv_name]
		
		
	try: 
		devnull = os.open('/dev/null',os.O_WRONLY)
		res = subprocess.check_output(['lvremove','-f', '%s/%s' % (VG_NAME,lv_name)],stderr=devnull, stdin=devnull)
		os.close(devnull)
		
	except:
		raise

def update_volume(lv_name,size,units='g'):
	"""lv_name - name of lvm volume without hostgroup
	size - expected size of volume in future"""
	vg = get_vg(units)
	pvs = get_pvs(units)
	lv = get_lv(lv_name,units)
	if lv == None:
		# create
		# mame misto na vg?
		if vg[2] < size:
			raise IOError((28,'No space left in volume group'))
		
		stripe = False
		if len(pvs) > 1:
			# budeme stripovat?
			# TODO - zjistit, jestli je na vsech PV dostatek mista
			stripe = True
		devnull = os.open('/dev/null',os.O_WRONLY)
		if stripe:
			print 'lvcreate - striped'
			subprocess.check_output(['lvcreate','-n',lv_name,'-L%d%s' % (size,units),'-i%d' % (len(pvs),),'-I%s' % (STRIPE_SIZE,),VG_NAME],stderr=devnull, stdin=devnull)
			
		else:
			print 'lvcreate'
			subprocess.check_output(['lvcreate','-n',lv_name,'-L%d%s' % (size,units),VG_NAME],stderr=devnull, stdin=devnull)
		
		os.close(devnull)
		return True
		
	else: # update
		if(size > lv[1] - 0.05 and size < lv[1] + 0.05): # uz tu je a je spravne veliky
			return True
		
		elif lv[1] > size:
			raise Exception('Unable to resize down!')
		
		else: # resize nahoru
			if vg[2] < (size-lv[1]):
				raise IOError((28,'No space left in volume group'))
			
			devnull = os.open('/dev/null',os.O_WRONLY)
			subprocess.check_output(['lvresize','-L%d%s' % (size,units),'%s/%s' % (VG_NAME,lv_name)],stderr=devnull, stdin=devnull)
			os.close(devnull)
			return True
	
	raise Exception('WTF? should not be here')
	
def create_snapshot(lv_name,size=50):
	"""docstring for create_snapshot"""
	# vytvori snapshot z lv disku
	# vstup: povinne lv_name, nepovinne size v GiB (default 50 GiB)
	# vystup: nazev snapshotu
	disk_path = '/dev/%s/%s' % (VG_NAME,lv_name)
	snap_name = '%s-snap-%s' % (lv_name,''.join(random.choice(string.ascii_lowercase + string.digits) for x in range(10)))
	try: 
		devnull = os.open('/dev/null',os.O_WRONLY)
		subprocess.check_output(['lvcreate','-n',snap_name,'-L%sGB' % size,'-s',disk_path],stderr=devnull, stdin=devnull)
		os.close(devnull)
	except: 
		raise
	
	with c2padmin.config.node_config_lock:
		c2padmin.config.node_config['active_snapshots'][snap_name] = time.time()
		
	return snap_name
	
