from subprocess import Popen,PIPE,check_output
from select import select
from binascii import unhexlify
from socket import inet_aton,inet_ntoa
import util,network,logger,config
import time,re,os
import c2padmin

logger = logger.getLogger(__name__)

RO_IFACE=0
RO_DST=1
RO_GW=2
RO_METRIC=6
RO_MASK=7

class OpenVPN(object):
	"""docstring for OpenVPN"""
	def __init__(self, remote):
		self._remote = remote
		self._vpn_process = None
		self._interface = None
		self._local_ip = None
		self._peer_ip = None
		self._remote_ip = None
		self._connected = False
		self._gre_interface = 'vpntun'
	
	def get_remote_ip(self):
		"""docstring for get_remote_ip"""
		first_line = True
		for line in open('/proc/net/route'):
			if first_line:
				first_line = False
				continue
			route = line.strip().split()
			if route[RO_IFACE] == self._interface and unhexlify(route[RO_GW])[::-1] == inet_aton(self._peer_ip):
				self._remote_ip = inet_ntoa(unhexlify(route[RO_DST])[::-1])
				return True
		return False
	
	def setup_gre_interface(self):
		if os.system('ip link sh dev {}'.format(self._gre_interface)) == 0:
			os.system('ip link del dev {}'.format(self._gre_interface))
		os.system('ip link add dev {} type gretap local {} remote {}'.format(self._gre_interface,self._local_ip,self._remote_ip))
		os.system('ip link set up dev {}'.format(self._gre_interface))
		return True
		
	def cleanup(self):
		"""docstring for cleanup"""
		if os.system('ip link sh dev {}'.format(self._gre_interface)) == 0:
			os.system('ip link del dev {}'.format(self._gre_interface))
	
	
	def vpn_process(self):
		args = [
			'openvpn',
			'--tls-client',
			'--port','1194',
			'--proto','udp',
			'--remote',self._remote,
			'--ca','/etc/c2padmin/cert/ca_bundle.crt',
			'--cert','/etc/c2padmin/cert/vpn.crt',
			'--key','/etc/c2padmin/cert/vpn.key',
			'--dev','tun',
			'--tun-mtu','2000',
			'--fragment','1300',
			'--mssfix',
			'--keepalive','10','120',
			'--pull'
		]
		self._vpn_process = Popen(args,stdin=PIPE,stdout=PIPE,stderr=PIPE, close_fds=True)
		try:
			while True:
				if config.node_config['roles']['vpn'] != 'master' or config.main_thread_signal != None:
					self._vpn_process.terminate()
					self._vpn_process.wait()
					self.cleanup()
					return
				rlist,wlist,xlist = select([self._vpn_process.stdout,self._vpn_process.stderr],[],[],0.1)
				for s in rlist:
					line = s.readline()
					if s == self._vpn_process.stdout:
						mymatch = re.match('.*ip link set dev ([a-z0-9]+) up',line)
						if mymatch != None:
							self._interface = mymatch.group(1)
							continue
						mymatch = re.match('.*ip addr add dev ([a-z0-9]+) local ([0-9\.]+) peer ([0-9\.]+)',line)
						if mymatch != None:
							if self._interface != mymatch.group(1):
								logger.error('WTF! VPN interface nesedi s prechozim radkem! puvodni:{} novy:{}'.format(self._interface,mymatch.group(1)))
							self._local_ip = mymatch.group(2)
							self._peer_ip = mymatch.group(3)
							continue
						mymatch = re.match('.*Initialization Sequence Completed',line)
						if mymatch != None and self._interface != None and self._local_ip != None and self._peer_ip != None:
							if self.get_remote_ip():
								self.setup_gre_interface()
								self._connected = True
								network.bridge_admin.vpn_active = True
								network.bridge_admin.apply_config()
								logger.info('Connected on interface {} local {} peer {} remote {}'.format(self._interface,self._local_ip,self._peer_ip, self._remote_ip))
							else:
								logger.error('Unable to get remote IP')
							continue
						print 'stdout: {}'.format(line)
					else:
						logger.error(line)
				retcode = self._vpn_process.poll()
				if retcode != None:
					with config.node_config_lock:
						config.node_config['roles']['vpn'] = 'slave'
					self.cleanup()
					return retcode
				time.sleep(0.1)
				
		except Exception as ex:
			c2padmin.dispatcher.unexpectly_end(ex)
	# TODO - vypnout VPN, kdyz uz nejsem master
	# TODO - uklidit po sobe, kdyz VPN skonci