import os, subprocess, stat, mmap, stat, fcntl, struct, time
import __builtin__
try:
	import dioutilc
	dioutilc_available = True
except:
	dioutilc_available = False


IOCTL_BLKGETSIZE64 = 0x80081272
IOCTL_BLKDISCARD = 0x1277
IOCTL_BLKDISCARDZEROES = 0x127c

IOSTAT_RIOS = 0
IOSTAT_RMERGES = 1
IOSTAT_RSECTORS = 2
IOSTAT_RTICKS = 3
IOSTAT_WIOS = 4
IOSTAT_WMERGES = 5
IOSTAT_WSECTORS = 6
IOSTAT_WTICKS = 7
IOSTAT_INFLIGHT = 8
IOSTAT_TICKS = 9
IOSTAT_TIQ = 10
IOSTAT_TS = 11

def open(*args,**kwargs):
	"""docstring for open"""
	ret = DioFile()
	ret.open(*args,**kwargs)
	return ret

class DiskStats:
	def __init__(self):
		self.devices = []
		self.stats = {'totals':{}}
		for d in os.listdir('/sys/block'):
			if d[0:2] == 'sd' and os.path.exists(os.path.join('/sys/block',d,'stat')):
				self.devices.append(d)
				self.stats[d] = {
					'raw': [],
					'rios': None,
					'wios': None,
					'rmerges': None,
					'wmerges': None,
					'rwait': None,
					'wwait': None
				}
		self.get_stats()
				
		
	def get_stats(self):
		for d in self.devices:
			try:
				fh = __builtin__.open('/sys/block/%s/stat' % (d,))
				self.stats[d]['raw'].insert(0,map(lambda x: int(x),fh.read().split())+[time.time()])
				fh.close()
			except:
				raise
				pass
			
			while len(self.stats[d]['raw']) > 2:
				self.stats[d]['raw'].pop()
			if len(self.stats[d]['raw']) > 1:
				timediff = self.stats[d]['raw'][0][IOSTAT_TS] - self.stats[d]['raw'][1][IOSTAT_TS]
				if timediff > 0:
					self.stats[d]['rios'] = float(self.stats[d]['raw'][0][IOSTAT_RIOS] - self.stats[d]['raw'][1][IOSTAT_RIOS])/timediff
					self.stats[d]['wios'] = float(self.stats[d]['raw'][0][IOSTAT_WIOS] - self.stats[d]['raw'][1][IOSTAT_WIOS])/timediff
					self.stats[d]['rmerges'] = float(self.stats[d]['raw'][0][IOSTAT_RMERGES] - self.stats[d]['raw'][1][IOSTAT_RMERGES])/timediff
					self.stats[d]['wmerges'] = float(self.stats[d]['raw'][0][IOSTAT_WMERGES] - self.stats[d]['raw'][1][IOSTAT_WMERGES])/timediff
					if self.stats[d]['rios'] > 0:
						self.stats[d]['rwait'] = float(self.stats[d]['raw'][0][IOSTAT_RTICKS] - self.stats[d]['raw'][1][IOSTAT_RTICKS])/(self.stats[d]['raw'][0][IOSTAT_RIOS] - self.stats[d]['raw'][1][IOSTAT_RIOS])
					else:
						self.stats[d]['rwait'] = 0.0
					if self.stats[d]['wios'] > 0:
						self.stats[d]['wwait'] = float(self.stats[d]['raw'][0][IOSTAT_WTICKS] - self.stats[d]['raw'][1][IOSTAT_WTICKS])/(self.stats[d]['raw'][0][IOSTAT_WIOS] - self.stats[d]['raw'][1][IOSTAT_WIOS])
					else:
						self.stats[d]['wwait'] = 0.0
		self.__calculate_totals()
		return self.stats
	
	def __calculate_totals(self):
		# totals
		rios = 0.0
		wios = 0.0
		rmerges = 0.0
		wmerges = 0.0
		# average	
		rwait = 0.0
		wwait = 0.0
		rwait_l = 0
		wwait_l = 0
		for d in self.devices:
			if self.stats[d]['rios'] != None:
				rios += self.stats[d]['rios']
			if self.stats[d]['wios'] != None:
				wios += self.stats[d]['wios']
			if self.stats[d]['rmerges'] != None:
				rmerges += self.stats[d]['rmerges']
			if self.stats[d]['wmerges'] != None:
				wmerges += self.stats[d]['wmerges']
			if self.stats[d]['rwait'] != None:
				rwait += self.stats[d]['rwait']
				rwait_l += 1
			if self.stats[d]['wwait'] != None:
				wwait += self.stats[d]['wwait']
				wwait_l += 1
		self.stats['totals']['rios'] = rios
		self.stats['totals']['wios'] = wios
		self.stats['totals']['rmerges'] = rmerges
		self.stats['totals']['wmerges'] = wmerges
		if rwait_l > 0:
			self.stats['totals']['rwait'] = rwait/rwait_l
		else:
			self.stats['totals']['rwait'] = 0
		if wwait_l > 0:
			self.stats['totals']['wwait'] = wwait/wwait_l
		else:
			self.stats['totals']['wwait'] = 0
			

class DioFile:
	# writen files will always be padded to FS block size
	def __init__(self, write_throttling=True, max_read_latency=7.0, max_write_latency=10.0, min_speed=5.0, disallow_discard = False):
		if not dioutilc_available:
			raise ValueError('dioutilc library not available')
		self.output = ""
		self.fh = -1
		self.pos = 0
		self.flags = 0
		self.writable = False
		self.readable = False
		self.eof = False
		self.stat = None
		self.bytecounter = 0
		self.disallow_discard = disallow_discard
		self.allows_discard = False
		self.discard_zeroes = False
		self.device_name = None
		self.discard_granularity = None
		self.discard_max_bytes = None
		self.discard_alignment = None
		
		self.write_throttling = write_throttling
		self.last_write_time = None
		self.writetimes = []
		if self.write_throttling:
			self.max_read_latency = float(max_read_latency)
			self.max_write_latency = float(max_write_latency)
			self.min_speed = float(min_speed)
			self.iostats = DiskStats()
	
	def open(self,filename, flags = 0):
		self.size = None
		if flags & os.O_RDWR:
			raise ValueError('Read/Write access not yet implemented!')
			self.writable = True
			self.readable = True
		elif flags & os.O_WRONLY:
			self.writable = True
			self.readable = False
		else:
			self.writable = False
			self.readable = True
		self.flags = flags
		self.fh = os.open(filename, flags | os.O_DIRECT)
		self.stat = os.fstat(self.fh)
		if stat.S_ISBLK(self.stat.st_mode):
			self.bs = 512
			self.size = os.lseek(self.fh, 0, 2)
			os.lseek(self.fh, 0, 0)
			major = os.major(self.stat.st_rdev)
			minor = os.minor(self.stat.st_rdev)
			try:
				for d in os.listdir('/sys/block/'):
					try:
						if __builtin__.open('/sys/block/%s/dev' % (d,)).read().strip() == '%d:%d' % (major,minor):
							self.device_name = d
							self.discard_granularity = int(__builtin__.open('/sys/block/%s/queue/discard_granularity' % (d,)).read().strip())
							self.discard_max_bytes = int(__builtin__.open('/sys/block/%s/queue/discard_max_bytes' % (d,)).read().strip())
							self.discard_alignment = int(__builtin__.open('/sys/block/%s/discard_alignment' % (d,)).read().strip())
							if self.discard_max_bytes > 0 and self.discard_granularity > 0:
								self.allows_discard = not self.disallow_discard
								
							self.discard_zero_block = chr(0)*self.discard_granularity
					except:
						pass
					if self.device_name != None:
						break
			except:
				pass
		else:
			self.bs = os.statvfs(filename).f_bsize

	def getsize(self):
		if self.size != None:
			return self.size
		else:
			return os.fstat(self.fh).st_size
	
	def get_bs(self):
		if self.fh == -1:
			raise ValueError('I/O operation on closed file')
		return self.bs
	def __check_cache(self):
		if self.bytecounter > (128<<20):
			myfh = os.open('/proc/sys/vm/drop_caches', os.O_WRONLY)
			os.write(myfh,'1')
			os.close(myfh)
			self.bytecounter = 0
	 
	def read(self, maxbytes):
		if maxbytes%self.bs != 0:
			raise ValueError('Size not divisible by chunksize')
		if not self.readable:
			raise OSError(9, 'Bad file descriptor')
		if self.fh == -1:
			raise ValueError('I/O operation on closed file')
		if len(self.output):
			raise BufferError('Outstanding writes! You must finish write with size in full FS blocks!')
		if self.eof:
			return ''
		buff = mmap.mmap(-1, maxbytes)
		bytes_read = dioutilc.readinto(self.fh, buff)
		buff.seek(0)
		ret = buff.read(bytes_read)
		buff.close()
		if bytes_read%self.bs != 0:
			# this means eof
			self.eof = True
		self.pos += bytes_read
		self.bytecounter += bytes_read
		self.__check_cache()
		return ret
		
	def write(self,data):
		datalen = len(data)
		while len(self.writetimes) > 10:
			self.writetimes.pop()
		if self.write_throttling:
			max_sleep_time = float(datalen)/(self.min_speed*1024*1024)
			ll = len(self.writetimes)
			if ll > 0:
				sleep_time = sum(self.writetimes)*datalen*2/ll
			else:
				sleep_time = 0.01
			st = self.iostats.get_stats()
			while (st['totals']['wwait'] > self.max_write_latency or st['totals']['rwait'] > self.max_read_latency) and self.last_write_time != None and (time.time() - self.last_write_time) < max_sleep_time:
				time.sleep(sleep_time)
				st = self.iostats.get_stats()
		self.last_write_time = time.time()
		self.last_datalen = datalen
		
		if self.discard_zeroes and self.allows_discard:
			assert self.discard_granularity > 0, 'Invalid discard granularity'
			# analyze data and send discard instead of write in case of zeros
			if (self.pos - self.discard_alignment) % self.discard_granularity or len(self.output) > 0:
				# not aligned
				# TODO write outstanding data to next aligned block
				ret = self._write(data)
			else:
				while len(data) >= self.discard_granularity:
					d = data[:self.discard_granularity]
					if d == self.discard_zero_block:
						self.zero(self.discard_granularity)

					else:
						ret = self._write(d)

					data = data[self.discard_granularity:]
				if len(data) > 0:
					ret = self._write(data)
		else:
			ret = self._write(data)
		if self.last_datalen != 0:
			self.writetimes.insert(0,(time.time() - self.last_write_time)/self.last_datalen)
			
		return ret
	
	def _write(self, data):
		if not self.writable:
			raise OSError(9, 'Bad file descriptor')
		if self.fh == -1:
			raise ValueError('I/O operation on closed file')
		self.output += data
		l = len(self.output)
		# get largest possible chunk of data that fits with bs
		chunksize = (l/self.bs)*self.bs
		if chunksize == 0:
			return
		buff = mmap.mmap(-1, chunksize)
		buff.write(self.output[:chunksize])
		bytes_writen = os.write(self.fh, buff)
		buff.close()
		self.output = self.output[chunksize:]
		self.pos += bytes_writen
		self.bytecounter += bytes_writen
		self.__check_cache()

	def seek(self, pos, whence = 0):
		if self.fh == -1:
			raise ValueError('I/O operation on closed file')
		if len(self.output):
			raise BufferError('Outstanding writes! You must write full FS blocks!')
		if whence == 0 and pos == self.pos:
			return self.pos
		newpos = os.lseek(self.fh, pos, whence)
		if newpos%self.bs != 0:
			os.lseek(self.fh, self.pos, 0)
			raise ValueError('Invalid file position')
		else:
			self.pos = newpos
		if self.pos < self.getsize():
			self.eof = False
		else:
			self.eof = True
		return self.pos

	def tell(self):
		if self.fh == -1:
			raise ValueError('I/O operation on closed file')
		return self.pos
	
	def __flush(self):
		l = len(self.output)
		if l > 0:
			assert l <= self.bs, 'Too much outstanding data'
			buff = mmap.mmap(-1, self.bs)
			buff.write(self.output)
			bytes_writen = os.write(self.fh, buff)
			buff.close()
			self.pos += bytes_writen
	
	def ioctl(self,req, fmt, *args):
	    buf = struct.pack(fmt, *(args or [0]))
	    buf = fcntl.ioctl(self.fh, req, buf)
	    return struct.unpack(fmt, buf)[0]

	def zero(self,size):
		assert size % self.bs == 0, 'Size not aligned'
		if self.allows_discard:
			try:
				self.ioctl(IOCTL_BLKDISCARDZEROES, 'I')
				self.ioctl(IOCTL_BLKDISCARD, 'LL', self.pos, size)
				self.seek(self.pos+size)
				return
			except:
				self.allows_discard = False
				pass
		# else
		return self._write(chr(0)*size)
		
	   

	def close(self):
		if self.fh == -1:
			raise ValueError('I/O operation on closed file')
		if self.writable:
			self.__flush()
		os.close(self.fh)
		self.fh = -1

